"""Machine Report module."""

import sys as _sys
import time as _time
import logging as _log

import numpy as _np
from matplotlib import pyplot as _plt

from ..search import PSSearch as _PSSearch
from ..clientarch import ClientArchiver as _CltArch, Time as _Time, \
    PVData as _PVData, PVDataSet as _PVDataSet
from ..sofb.csdev import ConstTLines as _SOFBCte
from .csdev import Const as _Cte
from .macschedule import MacScheduleData as _MacScheduleData
from .utils import interp1d_previous as _interp1d_previous


class MacReport:
    """Machine reports.

    Based on archiver data and machine schedule data.

    Reports:

    - usershift_progmd_time
        Programmed time[h].
    - usershift_delivd_time
        Delivered time[h], considering current above initial
        current*THOLD_FACTOR_USERSSBEAM.
    - usershift_extra_time
        Extra time[h].
    - usershift_total_time
        Total time(delivered + extra)[h].
    - usershift_progmd_count
        Number of shifts programmed.
    - usershift_current_average
        Current average.
    - usershift_current_stddev
        Current standard deviation.
    - usershift_current_beg_average
        Current average at the beginning of the shift.
    - usershift_current_beg_stddev
        Current st.dev. at the beginning of the shift.
    - usershift_current_end_average
        Current average at the end of the shift.
    - usershift_current_end_stddev
        Current st.dev. at the end of the shift.
    - usershift_total_failures_time
        Total failures time[h].
    - usershift_failures_count
        Number of failures.
    - usershift_beam_dump_count
        Number of beam dumps.
    - usershift_time_to_recover_average
        Average time took to recover from failures[h].
    - usershift_time_to_recover_stddev
        Time std.dev. took to recover from failures[h].
    - usershift_time_between_failures_average
        Average time between failure occurrences[h].
    - usershift_beam_reliability
        Beam reliability. Ratio between delivered and programmed time.
    - usershift_total_stable_beam_time
        Total stable beam time[h].
    - usershift_total_unstable_beam_time
        Total unstable beam time[h].
    - usershift_unstable_beam_count
        Number of unstable beam occurrences.
    - usershift_time_between_unstable_beams_average
        Average time between unstable beam occurrences[h].
    - usershift_relative_stable_beam_time
        Ratio between stable + delivered and programmed time.
    - usershift_total_injection_time
        Total injection shift[h].
    - usershift_injection_count
        Number of injections.
    - usershift_injection_time_average
        Average time used for injection[h].
        (usershift_injection_time/usershift_injection_count)
    - usershift_injection_time_stddev
        Standard deviation of time used for injection[h].

    - lsusage_total_time
        Total time of Light Source Usage.
    - lsusage_machinestudy_failures_time
        Failures time in Machine Study Shift.
    - lsusage_machinestudy_failures
        Relative failures time in Machine Study Shift.
    - lsusage_machinestudy_operational_time
        Operational time in Machine Study Shift.
    - lsusage_machinestudy_operational
        Relative operational time in Machine Study Shift.
    - lsusage_machinestudy_time
        Time in Machine Study Shift.
    - lsusage_machinestudy
        Relative time in Machine Study Shift.
    - lsusage_commissioning_failures_time
        Failures time in Commissioning Shift.
    - lsusage_commissioning_failures
        Relative failures time in Commissioning Shift.
    - lsusage_commissioning_operational_time
        Operational time in Commissioning Shift.
    - lsusage_commissioning_operational
        Relative operational time in Commissioning Shift.
    - lsusage_commissioning_time
        Time in Commissioning Shift.
    - lsusage_commissioning
        Relative time in Commissioning Shift.
    - lsusage_conditioning_failures_time
        Failures time in Conditioning Shift.
    - lsusage_conditioning_failures
        Relative failures time in Conditioning Shift.
    - lsusage_conditioning_operational_time
        Operational time in Conditioning Shift.
    - lsusage_conditioning_operational
        Relative operational time in Conditioning Shift.
    - lsusage_conditioning_time
        Time in Conditioning Shift.
    - lsusage_conditioning
        Relative time in Conditioning Shift.
    - lsusage_maintenance_failures_time
        Failures time in Maintenance Shift.
    - lsusage_maintenance_failures
        Relative failures time in Maintenance Shift.
    - lsusage_maintenance_operational_time
        Operational time in Maintenance Shift.
    - lsusage_maintenance_operational
        Relative operational time in Maintenance Shift.
    - lsusage_maintenance_time
        Time in Maintenance Shift.
    - lsusage_maintenance
        Relative time in Maintenance Shift.
    - lsusage_user_failures_time
        Failures time in Users Shift.
    - lsusage_user_failures
        Relative failures time in Users Shift.
    - lsusage_user_operational_time
        Operational time in Users Shift.
    - lsusage_user_operational
        Relative operational time in Users Shift.
    - lsusage_user_time
        Time in Users Shift.
    - lsusage_user
        Relative time in Users Shift.

    - current_machinestudy_singlebunch_average
        Current average for machine study single bunch shifts.
    - current_machinestudy_singlebunch_stddev
        Current standard deviation for machine study single bunch shifts.
    - current_machinestudy_singlebunch_time
        Time of machine study single bunch shifts.
    - current_machinestudy_multibunch_average
        Current average for machine study multi bunch shifts.
    - current_machinestudy_multibunch_stddev
        Current standard deviation for machine study multi bunch shifts.
    - current_machinestudy_multibunch_time
        Time of machine study multi bunch shifts.
    - current_machinestudy_total_average
        Current average for machine study shifts.
    - current_machinestudy_total_stddev
        Current standard deviation for machine study shifts.
    - current_machinestudy_total_time
        Time of machine study shifts.
    - current_commissioning_singlebunch_average
        Current average for single bunch commissioning shifts.
    - current_commissioning_singlebunch_stddev
        Current standard deviation for single bunch commissioning shifts.
    - current_commissioning_singlebunch_time
        Time of single bunch commissioning shifts.
    - current_commissioning_multibunch_average
        Current average for multi bunch commissioning shifts.
    - current_commissioning_multibunch_stddev
        Current standard deviation for multi bunch commissioning shifts.
    - current_commissioning_multibunch_time
        Time of multi bunch commissioning shifts.
    - current_commissioning_total_average
        Current average in commissioning shifts.
    - current_commissioning_total_stddev
        Current standard deviation in commissioning shifts.
    - current_commissioning_total_time
        Time of commissioning shifts.
    - current_conditioning_singlebunch_average
        Current average for single bunch conditioning shifts.
    - current_conditioning_singlebunch_stddev
        Current standard deviation for single bunch conditioning shifts.
    - current_conditioning_singlebunch_time
        Time of single bunch conditioning shifts.
    - current_conditioning_multibunch_average
        Current average for multi bunch conditioning shifts.
    - current_conditioning_multibunch_stddev
        Current standard deviation for multi bunch conditioning shifts.
    - current_conditioning_multibunch_time
        Time of multi bunch conditioning shifts.
    - current_conditioning_total_average
        Current average in conditioning shifts.
    - current_conditioning_total_stddev
        Current standard deviation in conditioning shifts.
    - current_conditioning_total_time
        Time of conditioning shifts.
    - current_user_singlebunch_average
        Current average for single bunch user shifts.
    - current_user_singlebunch_stddev
        Current standard deviation for single bunch user shifts.
    - current_user_singlebunch_time
        Time of single bunch user shifts.
    - current_user_multibunch_average
        Current average for multi bunch user shifts.
    - current_user_multibunch_stddev
        Current standard deviation for multi bunch user shifts.
    - current_user_multibunch_time
        Time of multi bunch user shifts.
    - current_user_total_average
        Current average in all user shifts.
    - current_user_total_stddev
        Current standard deviation in all user shifts.
    - current_user_total_time
        Time of all user shifts.
    - current_ebeam_singlebunch_average
        Current average for all single bunch shifts.
    - current_ebeam_singlebunch_stddev
        Current standard deviation for all single bunch shifts.
    - current_ebeam_singlebunch_time
        Time of all single bunch shifts.
    - current_ebeam_multibunch_average
        Current average for all multi bunch shifts.
    - current_ebeam_multibunch_stddev
        Current standard deviation for all multi bunch shifts.
    - current_ebeam_multibunch_time
        Time of all multi bunch shifts.
    - current_ebeam_total_average
        Average current considering the entire time in which
        there was any current above a threshold (THOLD_STOREDBEAM)
    - current_ebeam_total_stddev
        Current standard deviation considering the entire time in which
        there was any current above a threshold (THOLD_STOREDBEAM)
    - current_ebeam_total_time
        Time in which there was stored beam, for any
        current value above a threshold (THOLD_STOREDBEAM)
    """

    THOLD_STOREDBEAM = 0.008  # [mA]
    THOLD_FACTOR_USERSSBEAM = 0.5  # 50%
    QUERY_AVG_TIME = 60  # [s]

    def __init__(self, connector=None, logger=None):
        """Initialize object."""
        # client archiver connector
        self._connector = connector or _CltArch()

        # auxiliary logger
        self.logger = logger

        # create pv connectors
        self._init_connectors()

        # query data
        self._time_start = None
        self._time_stop = None

        # user shift stats
        self._usershift_progmd_time = None
        self._usershift_delivd_time = None
        self._usershift_extra_time = None
        self._usershift_total_time = None
        self._usershift_progmd_count = None
        self._usershift_current_average = None
        self._usershift_current_stddev = None
        self._usershift_current_beg_average = None
        self._usershift_current_beg_stddev = None
        self._usershift_current_end_average = None
        self._usershift_current_end_stddev = None
        self._usershift_total_failures_time = None
        self._usershift_failures_count = None
        self._usershift_beam_dump_count = None
        self._usershift_time_to_recover_average = None
        self._usershift_time_to_recover_stddev = None
        self._usershift_time_between_failures_average = None
        self._usershift_beam_reliability = None
        self._usershift_total_stable_beam_time = None
        self._usershift_total_unstable_beam_time = None
        self._usershift_unstable_beam_count = None
        self._usershift_time_between_unstable_beams_average = None
        self._usershift_relative_stable_beam_time = None
        self._usershift_total_injection_time = None
        self._usershift_injection_count = None
        self._usershift_injection_time_average = None
        self._usershift_injection_time_stddev = None

        # light source usage stats
        self._lsusage_total_time = None
        self._lsusage_machinestudy_failures_time = None
        self._lsusage_machinestudy_failures = None
        self._lsusage_machinestudy_operational_time = None
        self._lsusage_machinestudy_operational = None
        self._lsusage_machinestudy_time = None
        self._lsusage_machinestudy = None
        self._lsusage_commissioning_failures_time = None
        self._lsusage_commissioning_failures = None
        self._lsusage_commissioning_operational_time = None
        self._lsusage_commissioning_operational = None
        self._lsusage_commissioning_time = None
        self._lsusage_commissioning = None
        self._lsusage_conditioning_failures_time = None
        self._lsusage_conditioning_failures = None
        self._lsusage_conditioning_operational_time = None
        self._lsusage_conditioning_operational = None
        self._lsusage_conditioning_time = None
        self._lsusage_conditioning = None
        self._lsusage_maintenance_failures_time = None
        self._lsusage_maintenance_failures = None
        self._lsusage_maintenance_operational_time = None
        self._lsusage_maintenance_operational = None
        self._lsusage_maintenance_time = None
        self._lsusage_maintenance = None
        self._lsusage_user_failures_time = None
        self._lsusage_user_failures = None
        self._lsusage_user_operational_time = None
        self._lsusage_user_operational = None
        self._lsusage_user_time = None
        self._lsusage_user = None

        # stored current stats
        self._current_machinestudy_singlebunch_average = None
        self._current_machinestudy_singlebunch_stddev = None
        self._current_machinestudy_singlebunch_time = None
        self._current_machinestudy_multibunch_average = None
        self._current_machinestudy_multibunch_stddev = None
        self._current_machinestudy_multibunch_time = None
        self._current_machinestudy_total_average = None
        self._current_machinestudy_total_stddev = None
        self._current_machinestudy_total_time = None
        self._current_commissioning_singlebunch_average = None
        self._current_commissioning_singlebunch_stddev = None
        self._current_commissioning_singlebunch_time = None
        self._current_commissioning_multibunch_average = None
        self._current_commissioning_multibunch_stddev = None
        self._current_commissioning_multibunch_time = None
        self._current_commissioning_total_average = None
        self._current_commissioning_total_stddev = None
        self._current_commissioning_total_time = None
        self._current_conditioning_singlebunch_average = None
        self._current_conditioning_singlebunch_stddev = None
        self._current_conditioning_singlebunch_time = None
        self._current_conditioning_multibunch_average = None
        self._current_conditioning_multibunch_stddev = None
        self._current_conditioning_multibunch_time = None
        self._current_conditioning_total_average = None
        self._current_conditioning_total_stddev = None
        self._current_conditioning_total_time = None
        self._current_user_singlebunch_average = None
        self._current_user_singlebunch_stddev = None
        self._current_user_singlebunch_time = None
        self._current_user_multibunch_average = None
        self._current_user_multibunch_stddev = None
        self._current_user_multibunch_time = None
        self._current_user_total_average = None
        self._current_user_total_stddev = None
        self._current_user_total_time = None
        self._current_ebeam_singlebunch_average = None
        self._current_ebeam_singlebunch_stddev = None
        self._current_ebeam_singlebunch_time = None
        self._current_ebeam_multibunch_average = None
        self._current_ebeam_multibunch_stddev = None
        self._current_ebeam_multibunch_time = None
        self._current_ebeam_total_average = None
        self._current_ebeam_total_stddev = None
        self._current_ebeam_total_time = None

        # auxiliary data
        self._raw_data = None
        self._curr_times = None
        self._curr_values = None
        self._ps_fail_values = None
        self._gamblk_fail_values = None
        self._sofbloop_fail_values = None
        self._mps_fail_values = None
        self._inj_shift_values = None
        self._machinestudy_shift_values = None
        self._commissioning_shift_values = None
        self._conditioning_shift_values = None
        self._maintenance_shift_values = None
        self._user_shift_values = None
        self._user_shift_act_values = None
        self._user_shift_progmd_values = None
        self._user_shift_inicurr_values = None
        self._user_shift_delivd_values = None
        self._user_shift_stable_values = None
        self._is_stored_total = None
        self._is_stored_users = None
        self._singlebunch_values = None
        self._multibunch_values = None
        self._failures_users = None
        self._distortions_users = None

    @property
    def connector(self):
        """Client archiver connector."""
        return self._connector

    @property
    def logger(self):
        """Logger."""
        return self._logger

    @logger.setter
    def logger(self, new_logger):
        self._logger = new_logger
        self._logger_message = ''
        if not new_logger:
            _log.basicConfig(format='%(asctime)s | %(message)s',
                             datefmt='%F %T', level=_log.INFO,
                             stream=_sys.stdout)

    @property
    def timestamp_start(self):
        """Query interval start timestamp."""
        if not self._time_start:
            return None
        return self._time_start.timestamp()

    @timestamp_start.setter
    def timestamp_start(self, new_timestamp):
        if not isinstance(new_timestamp, (float, int)):
            raise TypeError('expected argument of type float or int')
        self._time_start = _Time(timestamp=new_timestamp)

    @property
    def time_start(self):
        """Time start."""
        return self._time_start

    @time_start.setter
    def time_start(self, new_time):
        if not isinstance(new_time, _Time):
            raise TypeError('expected argument of type Time')
        self._time_start = new_time

    @property
    def timestamp_stop(self):
        """Query interval stop timestamp."""
        if not self._time_stop:
            return None
        return self._time_stop.timestamp()

    @timestamp_stop.setter
    def timestamp_stop(self, new_timestamp):
        if not isinstance(new_timestamp, (float, int)):
            raise TypeError('expected argument of type float or int')
        self._time_stop = _Time(timestamp=new_timestamp)

    @property
    def time_stop(self):
        """Time stop."""
        return self._time_stop

    @time_stop.setter
    def time_stop(self, new_time):
        if not isinstance(new_time, _Time):
            raise TypeError('expected argument of type Time')
        self._time_stop = new_time

    # user shift stats

    @property
    def usershift_progmd_time(self):
        """Programmed time [h]."""
        return self._conv_sec_2_hour(self._usershift_progmd_time)

    @property
    def usershift_delivd_time(self):
        """Delivered time [h]."""
        return self._conv_sec_2_hour(self._usershift_delivd_time)

    @property
    def usershift_extra_time(self):
        """Extra time [h]."""
        return self._conv_sec_2_hour(self._usershift_extra_time)

    @property
    def usershift_total_time(self):
        """Total time (delivered + extra) [h]."""
        return self._conv_sec_2_hour(self._usershift_total_time)

    @property
    def usershift_current_average(self):
        """Current average."""
        return self._usershift_current_average

    @property
    def usershift_current_stddev(self):
        """Current standard deviation."""
        return self._usershift_current_stddev

    @property
    def usershift_current_beg_average(self):
        """Current average at the beginning of the shift."""
        return self._usershift_current_beg_average

    @property
    def usershift_current_beg_stddev(self):
        """Current st.dev. at the beginning of the shift."""
        return self._usershift_current_beg_stddev

    @property
    def usershift_current_end_average(self):
        """Current average at the end of the shift."""
        return self._usershift_current_end_average

    @property
    def usershift_current_end_stddev(self):
        """Current st.dev. at the end of the shift."""
        return self._usershift_current_end_stddev

    @property
    def usershift_progmd_count(self):
        """Number of programmed shifts."""
        return self._usershift_progmd_count

    @property
    def usershift_total_failures_time(self):
        """Total failures time [h]."""
        return self._conv_sec_2_hour(self._usershift_total_failures_time)

    @property
    def usershift_failures_count(self):
        """Number of failures."""
        return self._usershift_failures_count

    @property
    def usershift_beam_dump_count(self):
        """Number of beam dumps."""
        return self._usershift_beam_dump_count

    @property
    def usershift_time_to_recover_average(self):
        """Average time took to recover from failures [h]."""
        return self._conv_sec_2_hour(self._usershift_time_to_recover_average)

    @property
    def usershift_time_to_recover_stddev(self):
        """Time std.dev. took to recover from failures [h]."""
        return self._conv_sec_2_hour(self._usershift_time_to_recover_stddev)

    @property
    def usershift_time_between_failures_average(self):
        """Average time between failure occurrences [h]."""
        return self._conv_sec_2_hour(
            self._usershift_time_between_failures_average)

    @property
    def usershift_beam_reliability(self):
        """Beam reliability. Ratio between delivered and programmed time."""
        return self._usershift_beam_reliability

    @property
    def usershift_total_stable_beam_time(self):
        """Total stable beam time [h]."""
        return self._conv_sec_2_hour(self._usershift_total_stable_beam_time)

    @property
    def usershift_total_unstable_beam_time(self):
        """Total unstable beam time [h]."""
        return self._conv_sec_2_hour(self._usershift_total_unstable_beam_time)

    @property
    def usershift_unstable_beam_count(self):
        """Number of unstable beam occurrences."""
        return self._usershift_unstable_beam_count

    @property
    def usershift_time_between_unstable_beams_average(self):
        """Average time between unstable beam occurrences [h]."""
        return self._conv_sec_2_hour(
            self._usershift_time_between_unstable_beams_average)

    @property
    def usershift_relative_stable_beam_time(self):
        """Ratio between stable + delivered and programmed time."""
        return self._usershift_relative_stable_beam_time

    @property
    def usershift_total_injection_time(self):
        """Total injection shift [h]."""
        return self._conv_sec_2_hour(self._usershift_total_injection_time)

    @property
    def usershift_injection_count(self):
        """Number of injections."""
        return self._usershift_injection_count

    @property
    def usershift_injection_time_average(self):
        """Average time used for injection [h]."""
        return self._conv_sec_2_hour(
            self._usershift_injection_time_average)

    @property
    def usershift_injection_time_stddev(self):
        """Standard deviation of time used for injection [h]."""
        return self._conv_sec_2_hour(self._usershift_injection_time_stddev)

    # light source usage stats

    @property
    def lsusage_total_time(self):
        """Total time of Light Source Usage."""
        return self._conv_sec_2_hour(self._lsusage_total_time)

    @property
    def lsusage_machinestudy_failures_time(self):
        """Failures time in Machine Study Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_machinestudy_failures_time)

    @property
    def lsusage_machinestudy_failures(self):
        """Relative failures time in Machine Study Shift."""
        return self._lsusage_machinestudy_failures

    @property
    def lsusage_machinestudy_operational_time(self):
        """Operational time in Machine Study Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_machinestudy_operational_time)

    @property
    def lsusage_machinestudy_operational(self):
        """Relative operational time in Machine Study Shift."""
        return self._lsusage_machinestudy_operational

    @property
    def lsusage_machinestudy_time(self):
        """Time in Machine Study Shift."""
        return self._conv_sec_2_hour(self._lsusage_machinestudy_time)

    @property
    def lsusage_machinestudy(self):
        """Relative time in Machine Study Shift."""
        return self._lsusage_machinestudy

    @property
    def lsusage_commissioning_failures_time(self):
        """Failures time in Commissioning Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_commissioning_failures_time)

    @property
    def lsusage_commissioning_failures(self):
        """Relative failures time in Commissioning Shift."""
        return self._lsusage_commissioning_failures

    @property
    def lsusage_commissioning_operational_time(self):
        """Operational time in Commissioning Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_commissioning_operational_time)

    @property
    def lsusage_commissioning_operational(self):
        """Relative operational time in Commissioning Shift."""
        return self._lsusage_commissioning_operational

    @property
    def lsusage_commissioning_time(self):
        """Time in Commissioning Shift."""
        return self._conv_sec_2_hour(self._lsusage_commissioning_time)

    @property
    def lsusage_commissioning(self):
        """Relative time in Commissioning Shift."""
        return self._lsusage_commissioning

    @property
    def lsusage_conditioning_failures_time(self):
        """Failures time in Conditioning Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_conditioning_failures_time)

    @property
    def lsusage_conditioning_failures(self):
        """Relative failures time in Conditioning Shift."""
        return self._lsusage_conditioning_failures

    @property
    def lsusage_conditioning_operational_time(self):
        """Operational time in Conditioning Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_conditioning_operational_time)

    @property
    def lsusage_conditioning_operational(self):
        """Relative operational time in Conditioning Shift."""
        return self._lsusage_conditioning_operational

    @property
    def lsusage_conditioning_time(self):
        """Time in Conditioning Shift."""
        return self._conv_sec_2_hour(self._lsusage_conditioning_time)

    @property
    def lsusage_conditioning(self):
        """Relative time in Conditioning Shift."""
        return self._lsusage_conditioning

    @property
    def lsusage_maintenance_failures_time(self):
        """Failures time in Maintenance Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_maintenance_failures_time)

    @property
    def lsusage_maintenance_failures(self):
        """Relative failures time in Maintenance Shift."""
        return self._lsusage_maintenance_failures

    @property
    def lsusage_maintenance_operational_time(self):
        """Operational time in Maintenance Shift."""
        return self._conv_sec_2_hour(
            self._lsusage_maintenance_operational_time)

    @property
    def lsusage_maintenance_operational(self):
        """Relative operational time in Maintenance Shift."""
        return self._lsusage_maintenance_operational

    @property
    def lsusage_maintenance_time(self):
        """Time in Maintenance Shift."""
        return self._conv_sec_2_hour(self._lsusage_maintenance_time)

    @property
    def lsusage_maintenance(self):
        """Relative time in Maintenance Shift."""
        return self._lsusage_maintenance

    @property
    def lsusage_user_failures_time(self):
        """Failures time in Users Shift."""
        return self._conv_sec_2_hour(self._lsusage_user_failures_time)

    @property
    def lsusage_user_failures(self):
        """Relative failures time in Users Shift."""
        return self._lsusage_user_failures

    @property
    def lsusage_user_operational_time(self):
        """Operational time in Users Shift."""
        return self._conv_sec_2_hour(self._lsusage_user_operational_time)

    @property
    def lsusage_user_operational(self):
        """Relative operational time in Users Shift."""
        return self._lsusage_user_operational

    @property
    def lsusage_user_time(self):
        """Time in Users Shift."""
        return self._conv_sec_2_hour(self._lsusage_user_time)

    @property
    def lsusage_user(self):
        """Relative time in Users Shift."""
        return self._lsusage_user

    # stored current stats

    @property
    def current_machinestudy_singlebunch_average(self):
        """Current average for machine study single bunch shifts."""
        return self._current_machinestudy_singlebunch_average

    @property
    def current_machinestudy_singlebunch_stddev(self):
        """Current standard deviation for machine study single bunch shifts."""
        return self._current_machinestudy_singlebunch_stddev

    @property
    def current_machinestudy_singlebunch_time(self):
        """Time of machine study single bunch shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(
            self._current_machinestudy_singlebunch_time)

    @property
    def current_machinestudy_multibunch_average(self):
        """Current average for machine study multi bunch shifts."""
        return self._current_machinestudy_multibunch_average

    @property
    def current_machinestudy_multibunch_stddev(self):
        """Current standard deviation for machine study multi bunch shifts."""
        return self._current_machinestudy_multibunch_stddev

    @property
    def current_machinestudy_multibunch_time(self):
        """Time of machine study multi bunch shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(
            self._current_machinestudy_multibunch_time)

    @property
    def current_machinestudy_total_average(self):
        """Current average for machine study shifts."""
        return self._current_machinestudy_total_average

    @property
    def current_machinestudy_total_stddev(self):
        """Current standard deviation for machine study shifts."""
        return self._current_machinestudy_total_stddev

    @property
    def current_machinestudy_total_time(self):
        """Time of machine study shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(self._current_machinestudy_total_time)

    @property
    def current_commissioning_singlebunch_average(self):
        """Current average for single bunch commissioning shifts."""
        return self._current_commissioning_singlebunch_average

    @property
    def current_commissioning_singlebunch_stddev(self):
        """Current standard deviation for single bunch commissioning shifts."""
        return self._current_commissioning_singlebunch_stddev

    @property
    def current_commissioning_singlebunch_time(self):
        """Time of single bunch commissioning shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(
            self._current_commissioning_singlebunch_time)

    @property
    def current_commissioning_multibunch_average(self):
        """Current average for multi bunch commissioning shifts."""
        return self._current_commissioning_multibunch_average

    @property
    def current_commissioning_multibunch_stddev(self):
        """Current standard deviation for multi bunch commissioning shifts."""
        return self._current_commissioning_multibunch_stddev

    @property
    def current_commissioning_multibunch_time(self):
        """Time of multi bunch commissioning shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(
            self._current_commissioning_multibunch_time)

    @property
    def current_commissioning_total_average(self):
        """Current average in commissioning shifts."""
        return self._current_commissioning_total_average

    @property
    def current_commissioning_total_stddev(self):
        """Current standard deviation in commissioning shifts."""
        return self._current_commissioning_total_stddev

    @property
    def current_commissioning_total_time(self):
        """Time of commissioning shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(
            self._current_commissioning_total_time)

    @property
    def current_conditioning_singlebunch_average(self):
        """Current average for single bunch conditioning shifts."""
        return self._current_conditioning_singlebunch_average

    @property
    def current_conditioning_singlebunch_stddev(self):
        """Current standard deviation for single bunch conditioning shifts."""
        return self._current_conditioning_singlebunch_stddev

    @property
    def current_conditioning_singlebunch_time(self):
        """Time of single bunch conditioning shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(
            self._current_conditioning_singlebunch_time)

    @property
    def current_conditioning_multibunch_average(self):
        """Current average for multi bunch conditioning shifts."""
        return self._current_conditioning_multibunch_average

    @property
    def current_conditioning_multibunch_stddev(self):
        """Current standard deviation for multi bunch conditioning shifts."""
        return self._current_conditioning_multibunch_stddev

    @property
    def current_conditioning_multibunch_time(self):
        """Time of multi bunch conditioning shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(
            self._current_conditioning_multibunch_time)

    @property
    def current_conditioning_total_average(self):
        """Current average in conditioning shifts."""
        return self._current_conditioning_total_average

    @property
    def current_conditioning_total_stddev(self):
        """Current standard deviation in conditioning shifts."""
        return self._current_conditioning_total_stddev

    @property
    def current_conditioning_total_time(self):
        """Time of conditioning shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(self._current_conditioning_total_time)

    @property
    def current_user_singlebunch_average(self):
        """Current average for single bunch user shifts."""
        return self._current_user_singlebunch_average

    @property
    def current_user_singlebunch_stddev(self):
        """Current standard deviation for single bunch user shifts."""
        return self._current_user_singlebunch_stddev

    @property
    def current_user_singlebunch_time(self):
        """Time of single bunch user shifts.
        Consider any stored current."""
        return self._conv_sec_2_hour(self._current_user_singlebunch_time)

    @property
    def current_user_multibunch_average(self):
        """Current average for multi bunch user shifts."""
        return self._current_user_multibunch_average

    @property
    def current_user_multibunch_stddev(self):
        """Current standard deviation for multi bunch user shifts."""
        return self._current_user_multibunch_stddev

    @property
    def current_user_multibunch_time(self):
        """Time of multi bunch user shifts. Consider any stored current."""
        return self._conv_sec_2_hour(self._current_user_multibunch_time)

    @property
    def current_user_total_average(self):
        """Current average in all user shifts."""
        return self._current_user_total_average

    @property
    def current_user_total_stddev(self):
        """Current standard deviation in all user shifts."""
        return self._current_user_total_stddev

    @property
    def current_user_total_time(self):
        """Time of user shifts. Consider any stored current."""
        return self._conv_sec_2_hour(self._current_user_total_time)

    @property
    def current_ebeam_singlebunch_average(self):
        """Current average for all single bunch shifts."""
        return self._current_ebeam_singlebunch_average

    @property
    def current_ebeam_singlebunch_stddev(self):
        """Current standard deviation for all single bunch shifts."""
        return self._current_ebeam_singlebunch_stddev

    @property
    def current_ebeam_singlebunch_time(self):
        """Time of all single bunch shifts. Consider any stored current."""
        return self._conv_sec_2_hour(self._current_ebeam_singlebunch_time)

    @property
    def current_ebeam_multibunch_average(self):
        """Current average for all multi bunch shifts."""
        return self._current_ebeam_multibunch_average

    @property
    def current_ebeam_multibunch_stddev(self):
        """Current standard deviation for all multi bunch shifts."""
        return self._current_ebeam_multibunch_stddev

    @property
    def current_ebeam_multibunch_time(self):
        """Time of all multi bunch shifts. Consider any stored current."""
        return self._conv_sec_2_hour(self._current_ebeam_multibunch_time)

    @property
    def current_ebeam_total_average(self):
        """Current average for all stored beam time."""
        return self._current_ebeam_total_average

    @property
    def current_ebeam_total_stddev(self):
        """Current standard deviation for all stored beam time."""
        return self._current_ebeam_total_stddev

    @property
    def current_ebeam_total_time(self):
        """Stored beam time [h]. Consider any stored current."""
        return self._conv_sec_2_hour(self._current_ebeam_total_time)

    @property
    def raw_data(self):
        """Shift data and failures details."""
        return self._raw_data

    def update(self):
        """Update."""
        if self._time_start >= self._time_stop:
            raise ValueError('Invalid time interval.')

        for pvn in self._pvnames:
            self._pvdata[pvn].time_start = self._time_start
            self._pvdata[pvn].time_stop = self._time_stop
        for pvds in self._pvdataset.values():
            pvds.time_start = self._time_start
            pvds.time_stop = self._time_stop

        self._update_log('Collecting archiver data...')

        log_msg = 'Query for {0} in archiver took {1:.3f}s'

        # current
        _t0 = _time.time()
        self._pvdata[self._current_pv].parallel_query_bin_interval = 60*60*6
        self._pvdata[self._current_pv].update(MacReport.QUERY_AVG_TIME)
        self._update_log(log_msg.format(self._current_pv, _time.time()-_t0))

        # macshift and sirius interlock
        for pvn in self._pvnames:
            if pvn == self._current_pv:
                continue
            _t0 = _time.time()
            self._pvdata[pvn].update(parallel=False)
            self._update_log(log_msg.format(pvn, _time.time()-_t0))

        # ps
        for group, pvdataset in self._pvdataset.items():
            _t0 = _time.time()
            pvdataset.update(parallel=False)
            self._update_log(log_msg.format(
                'SI PS '+group.capitalize(), _time.time()-_t0))

        self._compute_stats()

    def plot_raw_data(self):
        """Plot raw data for period timestamp_start to timestamp_stop."""
        if not self._raw_data:
            print('No data to display. Call update() to get data.')
            return

        datetimes = _np.array([_Time(t) for t in self._raw_data['Timestamp']])

        fig, axs = _plt.subplots(14, 1, sharex=True)
        fig.set_size_inches(9, 10)
        fig.subplots_adjust(top=0.96, left=0.08, bottom=0.05, right=0.96)
        axs[0].set_title('Raw data', fontsize=12)

        axs[0].xaxis.axis_date()
        axs[0].plot(
            datetimes, self._raw_data['Current'], '-',
            color='blue', label='Current')
        axs[0].legend(loc='upper left', fontsize=9)
        axs[0].grid()

        axs[1].xaxis.axis_date()
        axs[1].plot(
            datetimes, self._raw_data['UserShiftInitCurr'], '-',
            color='blue', label='User Shifts - Initial Current')
        axs[1].legend(loc='upper left', fontsize=9)
        axs[1].grid()

        axs[2].xaxis.axis_date()
        axs[2].plot(
            datetimes, self._raw_data['UserShiftProgmd'], '-',
            color='gold', label='User Shifts - Programmed')
        axs[2].legend(loc='upper left', fontsize=9)
        axs[2].grid()

        axs[3].xaxis.axis_date()
        axs[3].plot(
            datetimes, self._raw_data['UserShiftDelivd'], '-',
            color='gold', label='User Shifts - Delivered')
        axs[3].legend(loc='upper left', fontsize=9)
        axs[3].grid()

        axs[4].xaxis.axis_date()
        axs[4].plot(
            datetimes, self._raw_data['UserShiftStable'], '-',
            color='gold', label='User Shifts - Delivered Without Distortions')
        axs[4].legend(loc='upper left', fontsize=9)
        axs[4].grid()

        axs[5].xaxis.axis_date()
        axs[5].plot(
            datetimes, self._raw_data['UserShiftTotal'], '-',
            color='gold', label='User Shifts - Total')
        axs[5].legend(loc='upper left', fontsize=9)
        axs[5].grid()

        axs[6].xaxis.axis_date()
        axs[6].plot(
            datetimes, self._raw_data['Failures']['NoEBeam'], '-',
            color='red', label='Failures - NoEBeam')
        axs[6].legend(loc='upper left', fontsize=9)
        axs[6].grid()

        axs[7].xaxis.axis_date()
        axs[7].plot(
            datetimes, self._raw_data['GammaShutter'], '-',
            color='red', label='Failures - Gamma Shutter Closed')
        axs[7].legend(loc='upper left', fontsize=9)
        axs[7].grid()

        axs[8].xaxis.axis_date()
        axs[8].plot(
            datetimes, self._raw_data['Failures']['WrongShift'], '-',
            color='red', label='Failures - WrongShift')
        axs[8].legend(loc='upper left', fontsize=9)
        axs[8].grid()

        axs[9].xaxis.axis_date()
        axs[9].plot(
            datetimes, self._raw_data['Failures']['SubsystemsNOk'], '-',
            color='red', label='Failures - PS, RF and MPS')
        axs[9].legend(loc='upper left', fontsize=9)
        axs[9].grid()

        axs[10].xaxis.axis_date()
        axs[10].plot(
            datetimes, self._raw_data['Distortions']['SOFBLoop'], '-',
            color='orangered', label='Distortions - SOFB Loop Open')
        axs[10].legend(loc='upper left', fontsize=9)
        axs[10].grid()

        axs[11].xaxis.axis_date()
        axs[11].plot(
            datetimes, self._raw_data['Shift']['Injection'], '-',
            color='lightsalmon', label='Injection Shifts')
        axs[11].legend(loc='upper left', fontsize=9)
        axs[11].grid()

        shift2color = {
            'MachineStudy': ['MacStudy', 'skyblue'],
            'Commissioning': ['Commi', 'royalblue'],
            'Conditioning': ['Condit', 'orchid'],
            'Maintenance': ['Maint', 'green']}
        for shift, auxdata in shift2color.items():
            ydata = self._raw_data['Shift'][shift]

            axs[12].xaxis.axis_date()
            axs[12].plot(
                datetimes, ydata, '-',
                color=auxdata[1], label=auxdata[0])
        axs[12].legend(loc='upper left', ncol=4, fontsize=9)
        axs[12].set_ylim(0.0, 2.0)
        axs[12].grid()

        egmodes2color = {
            'MultiBunch': 'orangered', 'SingleBunch': 'orange'}
        for egmode, color in egmodes2color.items():
            ydata = self._raw_data['EgunModes'][egmode]

            axs[13].xaxis.axis_date()
            axs[13].plot(
                datetimes, ydata, '-',
                color=color, label=egmode)
        axs[13].legend(loc='upper left', ncol=2, fontsize=9)
        axs[13].set_ylim(0.0, 2.0)
        axs[13].grid()

        return fig

    def plot_progmd_vs_delivd_hours(self):
        """Plot programmed vs. delivered hours graph."""
        if not self._raw_data:
            print('No data to display. Call update() to get data.')
            return

        datetimes = [_Time(d) for d in self._raw_data['Timestamp']]

        dtimes = _np.diff(self._raw_data['Timestamp'])
        dtimes = _np.r_[dtimes, dtimes[-1]]
        dtimes = dtimes/60/60

        dtimes_users_progmd = dtimes*self._raw_data['UserShiftProgmd']
        cum_progmd = _np.cumsum(dtimes_users_progmd)
        dtimes_users_delivd = dtimes*self._raw_data['UserShiftDelivd']
        cum_deliv = _np.cumsum(dtimes_users_delivd)

        fig = _plt.figure()
        axs = _plt.gca()
        axs.xaxis.axis_date()
        axs.plot(datetimes, cum_progmd, '-', label='Programmed')
        axs.plot(datetimes, cum_deliv, '-', label='Delivered')
        axs.grid()
        axs.set_ylabel('Integrated Hours')
        _plt.legend(loc=4)
        _plt.title('Integrated User Hours')
        return fig

    # ----- auxiliary methods -----

    def _init_connectors(self):
        self._current_pv = 'SI-Glob:AP-CurrInfo:Current-Mon'
        self._macshift_pv = 'AS-Glob:AP-MachShift:Mode-Sts'
        self._egtrgen_pv = 'LI-01:EG-TriggerPS:enablereal'
        self._egpusel_pv = 'LI-01:EG-PulsePS:singleselstatus'
        self._injevt_pv = 'AS-RaMO:TI-EVG:InjectionEvt-Sts'
        self._gammashutt_pv = 'AS-Glob:PP-GammaShutter:Status-Mon'
        self._siintlk_pv = 'RA-RaSIA02:RF-IntlkCtrl:IntlkSirius-Mon'
        self._sisofbloop_pv = 'SI-Glob:AP-SOFB:LoopState-Sts'
        self._pvnames = [
            self._current_pv, self._macshift_pv,
            self._egtrgen_pv, self._egpusel_pv, self._injevt_pv,
            self._gammashutt_pv, self._siintlk_pv,
            self._sisofbloop_pv]

        self._pvdata = dict()
        self._pv2default = dict()
        for pvname in self._pvnames:
            self._pvdata[pvname] = _PVData(pvname, self._connector)
            self._pv2default[pvname] = 0.0 if pvname != self._macshift_pv\
                else _Cte.MachShift.Commissioning

        self._si_fams_psnames = _PSSearch.get_psnames(
            {'sec': 'SI', 'sub': 'Fam', 'dev': '(B|Q|S).*'})
        self._psgroup2psname = {'fams': self._si_fams_psnames}
        groups = {'corrs': '(CH|CV)',
                  'trims': 'Q(F|D|[1-4]).*',
                  'skews': 'QS'}
        subs = ['(0[1-5]).*', '(0[6-9]|10).*',
                '(1[1-5]).*', '(1[6-9]|20).*']
        for psdesc, psreg in groups.items():
            for idx, sub in enumerate(subs):
                key = psdesc+' ('+str(idx+1)+'/'+str(len(subs))+')'
                value = _PSSearch.get_psnames(
                    {'sec': 'SI', 'sub': sub, 'dev': psreg})
                self._psgroup2psname[key] = value

        self._pvdataset = dict()
        for group, psnames in self._psgroup2psname.items():
            pvnames = [psn+':DiagStatus-Mon' for psn in psnames]
            self._pvdataset[group] = _PVDataSet(pvnames, self._connector)
            for pvn in pvnames:
                self._pvdata[pvn] = self._pvdataset[group][pvn]
                self._pv2default[pvn] = 0.0

    def _compute_stats(self):
        self._raw_data = dict()

        # current data
        _curr_times, _curr_values = \
            self._get_pv_data('SI-Glob:AP-CurrInfo:Current-Mon')
        _curr_values[_curr_values < 0] = 0
        _curr_values[_curr_values > 500] = 0

        # resample current data, from 1 pt in 60s, to 1pt in 5s
        new_len = (len(_curr_times)-1)*12 + 1
        new_times = _np.linspace(_curr_times[0], _curr_times[-1], new_len)
        self._curr_times = new_times
        self._curr_values = _np.interp(
            new_times, _curr_times, _curr_values)

        self._raw_data['Timestamp'] = self._curr_times
        self._raw_data['Current'] = self._curr_values

        # ps status data
        psfail_all = [0] * len(self._curr_times)
        for psnames in self._psgroup2psname.values():
            for psn in psnames:
                psfail_times, psfail_values = \
                    self._get_pv_data(psn+':DiagStatus-Mon')
                psfail_values = _np.bitwise_and(  # disregard alarms
                    psfail_values.astype(int), 0b1101111)

                psfail = 1 * (_interp1d_previous(
                    psfail_times, psfail_values, self._curr_times) > 0)

                psfail_all = _np.logical_or(psfail_all, psfail)
        self._ps_fail_values = 1 * psfail_all

        gamblk_times, gamblk_values = \
            self._get_pv_data('AS-Glob:PP-GammaShutter:Status-Mon')
        self._gamblk_fail_values = _interp1d_previous(
            gamblk_times, gamblk_values, self._curr_times)
        self._raw_data['GammaShutter'] = self._gamblk_fail_values

        sofbloop_times, sofbloop_values = \
            self._get_pv_data('SI-Glob:AP-SOFB:LoopState-Sts')
        sofbloop_fail_rawvalues = _np.array(
            [1*(v == _SOFBCte.LoopState.Open) for v in sofbloop_values])
        self._sofbloop_fail_values = _interp1d_previous(
            sofbloop_times, sofbloop_fail_rawvalues, self._curr_times)

        # rf and mps status data
        siintlk_times, siintlk_values = \
            self._get_pv_data('RA-RaSIA02:RF-IntlkCtrl:IntlkSirius-Mon')
        self._mps_fail_values = _interp1d_previous(
            siintlk_times, siintlk_values, self._curr_times)

        # delivered shift data
        shift_times, shift_values = \
            self._get_pv_data('AS-Glob:AP-MachShift:Mode-Sts')

        self._raw_data['Shift'] = dict()

        inj_shift_values = _np.array(
            [1*(v == _Cte.MachShift.Injection) for v in shift_values])
        self._inj_shift_values = _interp1d_previous(
            shift_times, inj_shift_values, self._curr_times)
        self._raw_data['Shift']['Injection'] = self._inj_shift_values

        stdy_shift_values = _np.array(
            [1*(v == _Cte.MachShift.MachineStudy) for v in shift_values])
        self._machinestudy_shift_values = _interp1d_previous(
            shift_times, stdy_shift_values, self._curr_times)
        self._raw_data['Shift']['MachineStudy'] = \
            self._machinestudy_shift_values

        cmm_shift_values = _np.array(
            [1*(v == _Cte.MachShift.Commissioning) for v in shift_values])
        self._commissioning_shift_values = _interp1d_previous(
            shift_times, cmm_shift_values, self._curr_times)
        self._raw_data['Shift']['Commissioning'] = \
            self._commissioning_shift_values

        cdt_shift_values = _np.array(
            [1*(v == _Cte.MachShift.Conditioning) for v in shift_values])
        self._conditioning_shift_values = _interp1d_previous(
            shift_times, cdt_shift_values, self._curr_times)
        self._raw_data['Shift']['Conditioning'] = \
            self._conditioning_shift_values

        mtn_shift_values = _np.array(
            [1*(v == _Cte.MachShift.Maintenance) for v in shift_values])
        self._maintenance_shift_values = _interp1d_previous(
            shift_times, mtn_shift_values, self._curr_times)
        self._raw_data['Shift']['Maintenance'] = self._maintenance_shift_values

        user_shift_values = _np.array(
            [1*(v == _Cte.MachShift.Users) for v in shift_values])
        self._user_shift_values = _interp1d_previous(
            shift_times, user_shift_values, self._curr_times)

        # desired shift data
        _t0 = _time.time()
        self._user_shift_progmd_values = \
            _MacScheduleData.is_user_shift_programmed(
                timestamp=self._curr_times)
        self._user_shift_inicurr_values = \
            _MacScheduleData.get_initial_current_programmed(
                timestamp=self._curr_times)
        self._usershift_progmd_count = \
            _MacScheduleData.get_users_shift_count(
                self._curr_times[0], self._curr_times[-1])
        self._update_log(
            'Query for machine schedule data took {0:.3f}s'.format(
                _time.time()-_t0))
        self._raw_data['UserShiftProgmd'] = self._user_shift_progmd_values
        self._raw_data['UserShiftInitCurr'] = self._user_shift_inicurr_values

        # single/multi bunch mode data
        inj_ts, inj_vs = self._get_pv_data('AS-RaMO:TI-EVG:InjectionEvt-Sts')
        inj_vs = _interp1d_previous(inj_ts, inj_vs, self._curr_times)
        trig_ts, trig_vs = self._get_pv_data('LI-01:EG-TriggerPS:enablereal')
        trig_vs = _interp1d_previous(trig_ts, trig_vs, self._curr_times)
        sb_ts, sb_vs = self._get_pv_data('LI-01:EG-PulsePS:singleselstatus')
        sb_vs = _interp1d_previous(sb_ts, sb_vs, self._curr_times)
        idcs1 = _np.where(inj_vs*trig_vs*sb_vs)[0]
        mode_ts = self._curr_times[idcs1]
        mode_vs = [1]*len(idcs1)
        idcs2 = _np.where(inj_vs*trig_vs*1*_np.logical_not(sb_vs))[0]
        mode_ts = _np.r_[mode_ts, self._curr_times[idcs2]]
        mode_vs = _np.r_[mode_vs, [0]*len(idcs2)]
        ind = mode_ts.argsort()
        mode_ts, mode_vs = mode_ts[ind], mode_vs[ind]
        mode_vs = _interp1d_previous(mode_ts, mode_vs, self._curr_times)
        self._singlebunch_values = mode_vs
        self._multibunch_values = _np.logical_not(mode_vs)
        self._raw_data['EgunModes'] = dict()
        self._raw_data['EgunModes']['SingleBunch'] = self._singlebunch_values
        self._raw_data['EgunModes']['MultiBunch'] = self._multibunch_values

        # is stored data
        self._is_stored_total = self._curr_values > MacReport.THOLD_STOREDBEAM
        self._is_stored_users = self._curr_values >= \
            self._user_shift_inicurr_values*MacReport.THOLD_FACTOR_USERSSBEAM

        # auxiliary time vectors
        dtimes = _np.diff(self._curr_times)
        dtimes = _np.r_[dtimes, dtimes[-1]]

        dtimes_users_progmd = dtimes*self._user_shift_progmd_values
        dtimes_injection = dtimes*self._inj_shift_values
        dtimes_machinestudy = dtimes*self._machinestudy_shift_values
        dtimes_commissioning = dtimes*self._commissioning_shift_values
        dtimes_conditioning = dtimes*self._conditioning_shift_values
        dtimes_maintenance = dtimes*self._maintenance_shift_values

        # failures
        self._raw_data['Failures'] = dict()
        self._raw_data['Failures']['SubsystemsNOk'] = _np.logical_or(
            self._ps_fail_values, self._mps_fail_values)
        self._raw_data['Failures']['GammaShutter'] = \
            self._gamblk_fail_values.astype(int)
        self._raw_data['Failures']['NoEBeam'] = \
            _np.logical_not(self._is_stored_users)

        # # # ignore wrong shift failures shorter than 60s
        wrong_shift = \
            1 * ((self._user_shift_progmd_values-self._user_shift_values) > 0)
        ignore_wrong_shift = _np.zeros(wrong_shift.shape)
        for i, val in enumerate(wrong_shift):
            if i >= len(wrong_shift)-12:
                break
            if val == 1 and not _np.sum(wrong_shift[(i-12):(i+12)]) >= 12:
                ignore_wrong_shift[i] = 1
        consider_wrong_shift = wrong_shift - ignore_wrong_shift
        self._raw_data['Failures']['WrongShift'] = consider_wrong_shift

        self._failures_users = 1 * _np.logical_or.reduce(
            [value for value in self._raw_data['Failures'].values()]) * \
            self._user_shift_progmd_values
        dtimes_failures_users = dtimes*self._failures_users

        self._user_shift_delivd_values = self._user_shift_progmd_values * \
            _np.logical_not(self._failures_users)
        self._raw_data['UserShiftDelivd'] = self._user_shift_delivd_values
        dtimes_users_delivd = dtimes*self._user_shift_delivd_values

        # distortions
        self._raw_data['Distortions'] = dict()
        self._raw_data['Distortions']['SOFBLoop'] = \
            self._sofbloop_fail_values

        self._distortions_users = 1 * _np.logical_or.reduce(
            [value for value in self._raw_data['Distortions'].values()]) * \
            self._user_shift_delivd_values
        dtimes_distortions_users = dtimes*self._distortions_users

        self._user_shift_stable_values = self._user_shift_delivd_values * \
            _np.logical_not(self._distortions_users)
        self._raw_data['UserShiftStable'] = self._user_shift_stable_values
        dtimes_users_stable = dtimes*self._user_shift_stable_values

        # user total and extra shift
        self._user_shift_act_values = \
            self._user_shift_values*_np.logical_not(self._failures_users)
        self._raw_data['UserShiftTotal'] = self._user_shift_act_values
        dtimes_users_total = dtimes*self._user_shift_act_values
        dtimes_users_extra = dtimes_users_total*_np.logical_not(
            self._user_shift_progmd_values)

        # calculate stats

        # # beam for users stats
        # # # ----- users shift -----
        self._usershift_progmd_time = _np.sum(dtimes_users_progmd)

        self._usershift_delivd_time = _np.sum(dtimes_users_delivd)

        self._usershift_extra_time = _np.sum(dtimes_users_extra)

        self._usershift_total_time = _np.sum(dtimes_users_total)

        self._usershift_current_average, self._usershift_current_stddev = \
            self._calc_current_stats(dtimes_users_total)

        transit = _np.diff(self._user_shift_delivd_values)
        beg_idcs = _np.where(transit == 1)[0]
        end_idcs = _np.where(transit == -1)[0]

        if beg_idcs.size:
            beg_val = [i for i in range(beg_idcs.size-1) if
                       beg_idcs[i+1]-beg_idcs[i] > 15]
            beg_val += [beg_idcs.size-1]
            beg1, beg2 = beg_idcs[beg_val], beg_idcs[beg_val] + 15
            if beg2[-1] > self._user_shift_delivd_values.size-1:
                beg1.pop()
                beg2.pop()
            beg_val = [i for i in range(beg1.size) if not
                       any([beg1[i] < e < beg2[i] for e in end_idcs])]
            beg1, beg2 = beg1[beg_val], beg2[beg_val]
            stats_vals = [_np.mean(self._curr_values[beg1[i]:beg2[i]])
                          for i in range(beg1.size)]
            self._usershift_current_beg_average = _np.mean(stats_vals)
            self._usershift_current_beg_stddev = _np.std(stats_vals)
        else:
            self._usershift_current_beg_average = 0
            self._usershift_current_beg_stddev = 0

        if end_idcs.size:
            end_val = [0] + [i for i in range(end_idcs.size)
                             if end_idcs[i]-end_idcs[i-1] > 15]
            end1, end2 = end_idcs[end_val] - 15, end_idcs[end_val]
            if end1[0] < 0:
                end1.pop(0)
                end2.pop(0)
            end_val = [i for i in range(end1.size-1) if not
                       any([end1[i] < b < end2[i] for b in beg_idcs])]
            end1, end2 = end1[end_val], end2[end_val]
            stats_vals = [_np.mean(self._curr_values[end1[i]:end2[i]])
                          for i in range(end1.size)]
            self._usershift_current_end_average = _np.mean(stats_vals)
            self._usershift_current_end_stddev = _np.std(stats_vals)
        else:
            self._usershift_current_end_average = 0
            self._usershift_current_end_stddev = 0

        # # # ----- failures -----
        self._usershift_total_failures_time = _np.sum(dtimes_failures_users)

        beam_dump_values = _np.logical_not(
            self._raw_data['Failures']['WrongShift']) * \
            self._raw_data['Failures']['NoEBeam']
        self._usershift_beam_dump_count = _np.sum(
            _np.diff(beam_dump_values) > 0)

        ave, std, count = self._calc_interval_stats(
            self._failures_users, dtimes_failures_users)
        self._usershift_time_to_recover_average = ave
        self._usershift_time_to_recover_stddev = std
        self._usershift_failures_count = count

        self._usershift_time_between_failures_average = \
            _np.inf if not self._usershift_failures_count\
            else self._usershift_progmd_time/self._usershift_failures_count

        # # # ----- distortions -----
        self._usershift_total_unstable_beam_time = _np.sum(
            dtimes_distortions_users)

        _, _, count = self._calc_interval_stats(
            self._distortions_users, dtimes_distortions_users)
        self._usershift_unstable_beam_count = count

        self._usershift_time_between_unstable_beams_average = \
            _np.inf if not self._usershift_unstable_beam_count\
            else self._usershift_progmd_time / \
            self._usershift_unstable_beam_count

        self._usershift_total_stable_beam_time = _np.sum(dtimes_users_stable)
        self._usershift_relative_stable_beam_time = \
            0.0 if not self._usershift_progmd_time else 100 * \
            self._usershift_total_stable_beam_time/self._usershift_progmd_time

        # # # ----- reliability -----
        self._usershift_beam_reliability = \
            0.0 if not self._usershift_progmd_time else 100 * \
            self._usershift_delivd_time/self._usershift_progmd_time

        # # # ----- injection shift -----
        self._usershift_total_injection_time = _np.sum(dtimes_injection)

        ave, std, count = self._calc_interval_stats(
            self._inj_shift_values, dtimes_injection)
        self._usershift_injection_time_average = ave
        self._usershift_injection_time_stddev = std
        self._usershift_injection_count = count

        # # light source usage stats
        self._lsusage_machinestudy_time = _np.sum(dtimes_machinestudy)

        self._lsusage_commissioning_time = _np.sum(dtimes_commissioning)

        self._lsusage_conditioning_time = _np.sum(dtimes_conditioning)

        self._lsusage_maintenance_time = _np.sum(dtimes_maintenance)

        self._lsusage_user_time = self._usershift_progmd_time + \
            self._usershift_extra_time

        self._lsusage_total_time = \
            self._lsusage_machinestudy_time + \
            self._lsusage_commissioning_time + \
            self._lsusage_conditioning_time + \
            self._lsusage_maintenance_time + \
            self._lsusage_user_time

        self._lsusage_machinestudy_failures_time = _np.sum(
            dtimes_machinestudy*self._raw_data['Failures']['SubsystemsNOk'])

        self._lsusage_commissioning_failures_time = _np.sum(
            dtimes_commissioning*self._raw_data['Failures']['SubsystemsNOk'])

        self._lsusage_conditioning_failures_time = _np.sum(
            dtimes_conditioning*self._raw_data['Failures']['SubsystemsNOk'])

        self._lsusage_maintenance_failures_time = 0.0

        dtimes_failures_users_oper = dtimes*self._failures_users
        self._lsusage_user_failures_time = _np.sum(
            dtimes_failures_users_oper)

        for usage in ['machinestudy', 'commissioning', 'conditioning',
                      'maintenance', 'user']:
            fail_intvl = getattr(self, '_lsusage_'+usage+'_failures_time')
            total_intvl = getattr(self, '_lsusage_'+usage+'_time')
            oper_intvl = total_intvl - fail_intvl
            if total_intvl:
                setattr(self, '_lsusage_'+usage+'_operational_time',
                        oper_intvl)
                setattr(self, '_lsusage_'+usage+'_failures',
                        100*fail_intvl/total_intvl)
                setattr(self, '_lsusage_'+usage+'_operational',
                        100*oper_intvl/total_intvl)
            else:
                setattr(self, '_lsusage_'+usage+'_operational_time', 0.0)
                setattr(self, '_lsusage_'+usage+'_failures', 0.0)
                setattr(self, '_lsusage_'+usage+'_operational', 0.0)
            setattr(self, '_lsusage_'+usage,
                    100*total_intvl/self._lsusage_total_time)

        # # stored current stats
        for shifttype in ['machinestudy', 'commissioning', 'conditioning',
                          'user', 'ebeam']:
            for fillmode in ['singlebunch', 'multibunch', 'total']:

                select = self._is_stored_total
                if fillmode != 'total':
                    fillmode_values = getattr(self, '_'+fillmode+'_values')
                    select = select*fillmode_values
                if shifttype != 'ebeam':
                    shift_values = getattr(self, '_'+shifttype+'_shift_values')
                    select = select*shift_values

                dtimes_select = dtimes*select

                pname = '_current_'+shifttype+'_'+fillmode

                setattr(self, pname+'_time', _np.sum(dtimes_select))

                avg, sdv = self._calc_current_stats(dtimes_select)
                setattr(self, pname+'_average', avg)
                setattr(self, pname+'_stddev', sdv)

    def _get_pv_data(self, pvname):
        t_start = self._time_start.timestamp()
        data = self._pvdata[pvname]
        defv = self._pv2default[pvname]
        if data.timestamp is None:
            times = _np.array([t_start, ])
            values = _np.array([defv, ])
        else:
            times = _np.array(data.timestamp)
            values = _np.array(data.value)
            if times[0] > t_start + MacReport.QUERY_AVG_TIME:
                times = _np.r_[t_start, times]
                values = _np.r_[defv, values]
        return times, values

    def _calc_current_stats(self, dtimes):
        interval = _np.sum(dtimes)
        if not interval:
            average = 0.0
            stddev = 0.0
        else:
            average = _np.sum(self._curr_values*dtimes)/interval

            aux = (self._curr_values - average)
            stddev = _np.sqrt(_np.sum(aux*aux*dtimes)/interval)
        return average, stddev

    def _calc_interval_stats(self, values, dtimes):
        transit = _np.diff(values)
        beg_idcs = _np.where(transit == 1)[0]
        end_idcs = _np.where(transit == -1)[0]
        if beg_idcs.size and end_idcs.size:
            if beg_idcs[0] > end_idcs[0]:
                end_idcs = end_idcs[1:]
            if beg_idcs.size > end_idcs.size:
                end_idcs = _np.r_[end_idcs, values.size-1]
            stats_vals = [_np.sum(dtimes[beg_idcs[i]:end_idcs[i]])
                          for i in range(beg_idcs.size)]
            avg = _np.mean(stats_vals)
            std = _np.std(stats_vals)
            return avg, std, beg_idcs.size
        return 0, 0, 0

    def _conv_sec_2_hour(self, seconds):
        if seconds is None:
            return None
        return seconds/60/60

    def _update_log(self, message=''):
        self._logger_message = message
        if self._logger:
            self._logger.update(message)
        _log.info(message)

    # ----- magic methods -----

    def __getitem__(self, pvname):
        return self._pvdata[pvname]

    def __str__(self):
        ppties_userbeam = [
            ['usershift_progmd_time', 'h'],
            ['usershift_delivd_time', 'h'],
            ['usershift_total_time', 'h'],
            ['usershift_extra_time', 'h'],
            ['usershift_progmd_count', ''],
            ['usershift_current_average', 'mA'],
            ['usershift_current_stddev', 'mA'],
            ['usershift_current_beg_average', 'mA'],
            ['usershift_current_beg_stddev', 'mA'],
            ['usershift_current_end_average', 'mA'],
            ['usershift_current_end_stddev', 'mA'],
            ['usershift_total_failures_time', 'h'],
            ['usershift_failures_count', ''],
            ['usershift_beam_dump_count', ''],
            ['usershift_time_to_recover_average', 'h'],
            ['usershift_time_to_recover_stddev', 'h'],
            ['usershift_time_between_failures_average', 'h'],
            ['usershift_beam_reliability', '%'],
            ['usershift_total_stable_beam_time', 'h'],
            ['usershift_total_unstable_beam_time', 'h'],
            ['usershift_unstable_beam_count', ''],
            ['usershift_time_between_unstable_beams_average', 'h'],
            ['usershift_relative_stable_beam_time', '%'],
            ['usershift_total_injection_time', 'h'],
            ['usershift_injection_count', ''],
            ['usershift_injection_time_average', 'h'],
            ['usershift_injection_time_stddev', 'h'],
        ]
        ppties_lsusage = [
            ['lsusage_total_time', 'h'],
            ['lsusage_machinestudy_failures_time', 'h'],
            ['lsusage_machinestudy_failures', '%'],
            ['lsusage_machinestudy_operational_time', 'h'],
            ['lsusage_machinestudy_operational', '%'],
            ['lsusage_machinestudy_time', 'h'],
            ['lsusage_machinestudy', '%'],
            ['lsusage_commissioning_failures_time', 'h'],
            ['lsusage_commissioning_failures', '%'],
            ['lsusage_commissioning_operational_time', 'h'],
            ['lsusage_commissioning_operational', '%'],
            ['lsusage_commissioning_time', 'h'],
            ['lsusage_commissioning', '%'],
            ['lsusage_conditioning_failures_time', 'h'],
            ['lsusage_conditioning_failures', '%'],
            ['lsusage_conditioning_operational_time', 'h'],
            ['lsusage_conditioning_operational', '%'],
            ['lsusage_conditioning_time', 'h'],
            ['lsusage_conditioning', '%'],
            ['lsusage_maintenance_failures_time', 'h'],
            ['lsusage_maintenance_failures', '%'],
            ['lsusage_maintenance_operational_time', 'h'],
            ['lsusage_maintenance_operational', '%'],
            ['lsusage_maintenance_time', 'h'],
            ['lsusage_maintenance', '%'],
            ['lsusage_user_failures_time', 'h'],
            ['lsusage_user_failures', '%'],
            ['lsusage_user_operational_time', 'h'],
            ['lsusage_user_operational', '%'],
            ['lsusage_user_time', 'h'],
            ['lsusage_user', '%'],
        ]
        ppties_storedcurrent = [
            ['current_machinestudy_singlebunch_average', 'mA'],
            ['current_machinestudy_singlebunch_stddev', 'mA'],
            ['current_machinestudy_singlebunch_time', 'h'],
            ['current_machinestudy_multibunch_average', 'mA'],
            ['current_machinestudy_multibunch_stddev', 'mA'],
            ['current_machinestudy_multibunch_time', 'h'],
            ['current_machinestudy_total_average', 'mA'],
            ['current_machinestudy_total_stddev', 'mA'],
            ['current_machinestudy_total_time', 'h'],
            ['current_commissioning_singlebunch_average', 'mA'],
            ['current_commissioning_singlebunch_stddev', 'mA'],
            ['current_commissioning_singlebunch_time', 'h'],
            ['current_commissioning_multibunch_average', 'mA'],
            ['current_commissioning_multibunch_stddev', 'mA'],
            ['current_commissioning_multibunch_time', 'h'],
            ['current_commissioning_total_average', 'mA'],
            ['current_commissioning_total_stddev', 'mA'],
            ['current_commissioning_total_time', 'h'],
            ['current_conditioning_singlebunch_average', 'mA'],
            ['current_conditioning_singlebunch_stddev', 'mA'],
            ['current_conditioning_singlebunch_time', 'h'],
            ['current_conditioning_multibunch_average', 'mA'],
            ['current_conditioning_multibunch_stddev', 'mA'],
            ['current_conditioning_multibunch_time', 'h'],
            ['current_conditioning_total_average', 'mA'],
            ['current_conditioning_total_stddev', 'mA'],
            ['current_conditioning_total_time', 'h'],
            ['current_user_singlebunch_average', 'mA'],
            ['current_user_singlebunch_stddev', 'mA'],
            ['current_user_singlebunch_time', 'h'],
            ['current_user_multibunch_average', 'mA'],
            ['current_user_multibunch_stddev', 'mA'],
            ['current_user_multibunch_time', 'h'],
            ['current_user_total_average', 'mA'],
            ['current_user_total_stddev', 'mA'],
            ['current_user_total_time', 'h'],
            ['current_ebeam_singlebunch_average', 'mA'],
            ['current_ebeam_singlebunch_stddev', 'mA'],
            ['current_ebeam_singlebunch_time', 'h'],
            ['current_ebeam_multibunch_average', 'mA'],
            ['current_ebeam_multibunch_stddev', 'mA'],
            ['current_ebeam_multibunch_time', 'h'],
            ['current_ebeam_total_average', 'mA'],
            ['current_ebeam_total_stddev', 'mA'],
            ['current_ebeam_total_time', 'h'],
        ]
        rst = 'User Beam Statistics\n'
        for ppty, unit in ppties_userbeam:
            ppty_text = ppty + ('' if not unit else ' ('+unit+')')
            rst += '{0:>46s}: {1:>8.3f}\n'.format(
                ppty_text, getattr(self, ppty))
        rst += 'Light Source Usage Statistics\n'
        for ppty, unit in ppties_lsusage:
            ppty_text = ppty.split('lsusage_')[1] + \
                ('' if not unit else ' ('+unit+')')
            rst += '{0:>46s}: {1:>8.3f}\n'.format(
                ppty_text, getattr(self, ppty))
        rst += 'Stored Current Statistics\n'
        for ppty, unit in ppties_storedcurrent:
            ppty_text = ppty.split('current_')[1] + \
                ('' if not unit else ' ('+unit+')')
            rst += '{0:>46s}: {1:>8.3f}\n'.format(
                ppty_text, getattr(self, ppty))
        return rst



"""Main cycle module."""

import sys as _sys
import time as _time
import logging as _log
import threading as _thread
from concurrent.futures import ThreadPoolExecutor

from ..namesys import Filter as _Filter, SiriusPVName as _PVName
from ..search import PSSearch as _PSSearch
from ..epics import PV as _PV

from .conn import Timing, PSCycler, PSCyclerFBP, LinacPSCycler, FOFBPSCycler
from .bo_cycle_data import DEFAULT_RAMP_DURATION
from .util import Const as _Const, \
    get_trigger_by_psname as _get_trigger_by_psname

TIMEOUT_SLEEP = 0.1
TIMEOUT_CHECK = 20
TIMEOUT_CONN = 0.5


class CycleController:
    """Class to perform automated cycle procedure."""

    def __init__(self, cyclers=None, timing=None,
                 is_bo=False, ramp_config=None, logger=None,
                 isadv=False):
        """Initialize."""
        # initialize auxiliar variables
        self._mode = None
        self._include_sitrims = None
        self._not_ctrl_ti = None
        self._cycle_duration = 0
        self._aux_cyclers = dict()
        self._cycle_trims_duration = 0
        self._checks_result = dict()
        self._si_aux_triggers = list()

        # in case user wants to cycle SI power supplies individually,
        # not following the standard procedure
        self._isadv = isadv

        # in case cyclers are not set and user wants to cycle bo
        self._is_bo = is_bo
        self._ramp_config = ramp_config

        # cyclers
        self.cyclers = cyclers

        # timing connector
        self.timing = timing

        # logger
        self._logger_message = ''
        self.logger = logger

    # --- main parameter setters ---

    @property
    def cyclers(self):
        """Return current cyclers."""
        return self._cyclers

    @cyclers.setter
    def cyclers(self, new_cyclers):
        if new_cyclers:
            if not isinstance(new_cyclers, dict):
                raise TypeError("Input 'new_cyclers' has to be a dict!")
            psnames2filt = list(new_cyclers.keys())
            ps2cycle = self._filter_psnames(
                psnames2filt, {'sec': '(LI|TB|TS|SI)', 'dis': 'PS'})
            ps2ramp = self._filter_psnames(
                psnames2filt, {'sec': 'BO', 'dis': 'PS'})
            if ps2cycle and ps2ramp:
                raise Exception('Can not cycle Booster with other sectors!')
            self._mode = 'Ramp' if ps2ramp else 'Cycle'
        else:
            # create cyclers, if needed
            if self._is_bo:
                psnames = _PSSearch.get_psnames({'sec': 'BO', 'dis': 'PS'})
                self._mode = 'Ramp'
            else:
                psnames = _PSSearch.get_psnames(
                    {'sec': '(LI|TB|TS|SI)', 'dis': 'PS'})
                self._mode = 'Cycle'
            new_cyclers = dict()
            for name in psnames:
                if 'LI' in name:
                    new_cyclers[name] = LinacPSCycler(name)
                elif _PSSearch.conv_psname_2_psmodel(name) == 'FOFB_PS':
                    new_cyclers[name] = FOFBPSCycler(name)
                elif _PSSearch.conv_psname_2_psmodel(name) == 'FBP':
                    new_cyclers[name] = PSCyclerFBP(name, self._ramp_config)
                else:
                    new_cyclers[name] = PSCycler(name, self._ramp_config)
        self._cyclers = new_cyclers

        # define if we will perform default SI cycle procedure
        include_sifam = bool(_Filter.process_filters(
            self._cyclers.keys(), {'sec': 'SI', 'sub': 'Fam', 'dis': 'PS'}))
        self._include_sitrims = include_sifam and not self._isadv

        # define not_ctrl_ti variable
        ps_no_ti = _PSSearch.get_psnames({'sec': 'LI', 'dis': 'PS'})
        ps_no_ti.extend(
            _PSSearch.get_psnames({'sec': 'SI', 'dis': 'PS', 'dev': 'FC.*'}))
        self._not_ctrl_ti = not bool(set(self._cyclers.keys()) - set(ps_no_ti))

        # define triggers
        self._triggers = _get_trigger_by_psname(self._cyclers.keys())

        if self._include_sitrims:
            # trims psnames
            self.trimnames = _PSSearch.get_psnames(
                {'sec': 'SI', 'sub': '[0-2][0-9](M|C).*', 'dis': 'PS',
                 'dev': '(CH|CV|QS|QD.*|QF.*|Q[1-4])'})

            # trims triggers
            self._si_aux_triggers = [
                'SI-Glob:TI-Mags-Skews', 'SI-Glob:TI-Mags-Corrs',
                'SI-Glob:TI-Mags-QTrims']
            self._triggers.update(self._si_aux_triggers)

            # move CV-2 and QS of C2 to trims group, if they are in cyclers
            qs_c2 = _PSSearch.get_psnames(
                {'sec': 'SI', 'sub': '[0-2][0-9]C2', 'dis': 'PS',
                 'dev': 'QS'})
            cv2_c2 = _PSSearch.get_psnames(
                {'sec': 'SI', 'sub': '[0-2][0-9]C2', 'dis': 'PS',
                 'dev': 'CV', 'idx': '2'})
            for psn in qs_c2 + cv2_c2:
                if psn in self._cyclers.keys():
                    self._aux_cyclers[psn] = self._cyclers.pop(psn)
        else:
            self.trimnames = list()

        # define cycle duration
        duration = 0
        for psname in self._cyclers.keys():
            duration = max(
                duration, self._cyclers[psname].cycle_duration(self._mode))
        self._cycle_duration = duration

        # egun pv
        if 'LI-01:PS-Spect' in self._cyclers.keys():
            self._pv_egun = _PV('LI-01:EG-TriggerPS:enablereal',
                                connection_timeout=0.05)

    def create_trims_cyclers(self):
        """Create trims cyclers."""
        self._update_log('Creating trims connections...')
        for idx, psn in enumerate(self.trimnames):
            if idx % 5 == 4 or idx == len(self.trimnames)-1:
                self._update_log(
                    'Created connections of {0}/{1} trims'.format(
                        str(idx+1), str(len(self.trimnames))))
            if psn in self._aux_cyclers.keys():
                continue
            if _PSSearch.conv_psname_2_psmodel(psn) == 'FBP':
                self._aux_cyclers[psn] = PSCyclerFBP(psn, self._ramp_config)
            else:
                self._aux_cyclers[psn] = PSCycler(psn, self._ramp_config)

        # wait for connections
        self._update_log('Waiting for connections...')
        for cycler in self._aux_cyclers.values():
            cycler.wait_for_connection(0.1)

        # calculate trims cycle duration
        duration = 0
        for psname in self.trimnames:
            duration = max(
                duration, self._aux_cyclers[psname].cycle_duration(self._mode))
        self._cycle_trims_duration = duration

    @property
    def timing(self):
        """Return timing connector."""
        return self._timing

    @timing.setter
    def timing(self, new_timing):
        self._timing = new_timing if new_timing is not None else Timing()

    @property
    def logger(self):
        """Return current logger."""
        return self._logger

    @logger.setter
    def logger(self, new_log):
        self._logger = new_log
        if not new_log:
            _log.basicConfig(format='%(asctime)s | %(message)s',
                             datefmt='%F %T', level=_log.INFO,
                             stream=_sys.stdout)

    # --- properties ---

    @property
    def psnames(self):
        """Power supplies to cycle."""
        return list(self._cyclers.keys())

    @property
    def mode(self):
        """Mode."""
        return self._mode

    @property
    def save_timing_size(self):
        """Save timing initial state task size."""
        return 2

    @property
    def prepare_timing_size(self):
        """Prepare timing task size."""
        return 3

    @property
    def prepare_ps_sofbmode_size(self):
        """Prepare PS SOFBMode task size."""
        prepare_ps_size = 2*(len(self.psnames)+1)
        if self._include_sitrims:
            prepare_ps_size += 2*len(self.trimnames)
        return prepare_ps_size

    @property
    def prepare_ps_opmode_slowref_size(self):
        """Prepare PS OpMode SlowRef task size."""
        prepare_ps_size = 2*(len(self.psnames)+1)
        if self._include_sitrims:
            prepare_ps_size += 3*len(self.trimnames)
        return prepare_ps_size

    @property
    def prepare_ps_current_zero_size(self):
        """Prepare PS current zero task size."""
        prepare_ps_size = 2*(len(self.psnames)+1)
        if self._include_sitrims:
            prepare_ps_size += 3*len(self.trimnames)
        return prepare_ps_size

    @property
    def prepare_ps_params_size(self):
        """Prepare PS parameters task size."""
        prepare_ps_size = 2*(len(self.psnames)+1)
        if self._include_sitrims:
            prepare_ps_size += 3*len(self.trimnames)
        return prepare_ps_size

    @property
    def prepare_ps_opmode_cycle_size(self):
        """Prepare PS task size."""
        prepare_ps_size = 2*(len(self.psnames)+1)
        return prepare_ps_size

    @property
    def cycle_size(self):
        """Cycle task size."""
        cycle_size = (
            2 +  # check timing
            len(self.psnames)+3 +  # check params
            len(self.psnames) +    # check opmode
            2 +  # set and check triggers enable
            3+round(self._cycle_duration) +  # cycle
            len(self.psnames) +  # check final
            len(self.psnames)+2)  # reset subsystems
        return cycle_size

    @property
    def cycle_trims_size(self):
        """Cycle trims task size."""
        cycle_trims_size = (
            2*2 +  # check timing
            2*2*len(self.trimnames) +  # set and check params
            2*2*len(self.trimnames) +  # set and check opmode
            2*2 +  # set and check triggers enable
            2*(3+round(self._cycle_trims_duration)) +  # cycle
            2*len(self.trimnames))  # check final
        return cycle_trims_size

    @property
    def restore_timing_size(self):
        """Restore timing initial state task size."""
        return 2

    @property
    def save_timing_max_duration(self):
        """Save timing initial state task maximum duration."""
        return 10

    @property
    def prepare_timing_max_duration(self):
        """Prepare timing task maximum duration."""
        return 10

    @property
    def prepare_ps_sofbmode_max_duration(self):
        """Prepare PS SOFBMode task maximum duration."""
        prepare_ps_max_duration = 5 + TIMEOUT_CHECK
        if self._include_sitrims:
            prepare_ps_max_duration += TIMEOUT_CHECK
        return prepare_ps_max_duration

    @property
    def prepare_ps_opmode_slowref_max_duration(self):
        """Prepare PS OpMode SlowRef task maximum duration."""
        prepare_ps_max_duration = 10 + TIMEOUT_CHECK
        if self._include_sitrims:
            prepare_ps_max_duration += 3*TIMEOUT_CHECK
        return prepare_ps_max_duration

    @property
    def prepare_ps_current_zero_max_duration(self):
        """Prepare PS current zero task maximum duration."""
        prepare_ps_max_duration = 10 + TIMEOUT_CHECK
        if self._include_sitrims:
            prepare_ps_max_duration += 3*TIMEOUT_CHECK
        return prepare_ps_max_duration

    @property
    def prepare_ps_params_max_duration(self):
        """Prepare PS parameters task maximum duration."""
        prepare_ps_max_duration = 10 + TIMEOUT_CHECK
        if self._include_sitrims:
            prepare_ps_max_duration += 3*TIMEOUT_CHECK
        return prepare_ps_max_duration

    @property
    def prepare_ps_opmode_cycle_max_duration(self):
        """Prepare PS task maximum duration."""
        prepare_ps_max_duration = 10 + TIMEOUT_CHECK
        if self._include_sitrims:
            prepare_ps_max_duration += 3*TIMEOUT_CHECK
        return prepare_ps_max_duration

    @property
    def cycle_max_duration(self):
        """Cycle task maximum duration."""
        cycle_max_duration = (
            8 +  # check timing
            TIMEOUT_CHECK +  # check params
            TIMEOUT_CHECK +  # check opmode
            TIMEOUT_CHECK +  # set and check triggers enable
            60 +  # wait for timing trigger
            round(self._cycle_duration) +  # cycle
            TIMEOUT_CHECK +  # check final
            TIMEOUT_CHECK)  # set and check ps slowref
        return cycle_max_duration

    @property
    def cycle_trims_max_duration(self):
        """Cycle trims task maximum duration."""
        cycle_trims_max_duration = (
            8 +  # check timing
            2*50 +  # check params
            2*50 +  # set and check opmode
            2*TIMEOUT_CHECK +  # set and check triggers enable
            2*10 +  # wait for timing trigger
            2*round(self._cycle_trims_duration) +  # cycle
            2*TIMEOUT_CHECK +  # check final
            2*50)  # set and check ps slowref
        return cycle_trims_max_duration

    @property
    def restore_timing_max_duration(self):
        """Restore timing initial state task maximum duration."""
        return 10

    # --- auxiliary commands ---

    def config_pwrsupplies(self, ppty, psnames):
        """Prepare power supplies to cycle according to mode."""
        if ppty == 'opmode':
            psnames = {
                p for p in psnames if 'LI' not in p and
                not _PSSearch.conv_psname_2_psmodel(p) == 'FOFB_PS'}
        if not psnames:
            return True

        self._update_log('Preparing power supplies '+ppty+'...')
        with ThreadPoolExecutor(max_workers=100) as executor:
            for idx, psname in enumerate(psnames):
                cycler = self._get_cycler(psname)
                if ppty == 'parameters':
                    target = cycler.prepare
                elif ppty == 'opmode':
                    target = cycler.set_opmode_cycle
                executor.submit(target, self.mode)
                if idx % 5 == 4 or idx == len(psnames)-1:
                    self._update_log(
                        'Sent '+ppty+' preparation to {0}/{1}'.format(
                            str(idx+1), str(len(psnames))))

    def config_timing(self):
        """Prepare timing to cycle according to mode."""
        if self._not_ctrl_ti:
            return
        self._timing.turnoff(self._triggers)
        self._update_log('Preparing Timing...')
        self._timing.prepare(self.mode, self._triggers)
        self._update_log(done=True)

    def check_pwrsupplies(self, ppty, psnames, timeout=TIMEOUT_CHECK):
        """Check all power supplies according to mode."""
        if ppty == 'opmode':
            psnames = {
                p for p in psnames if 'LI' not in p and
                not _PSSearch.conv_psname_2_psmodel(p) == 'FOFB_PS'}
        if not psnames:
            return True

        self._update_log('Checking power supplies '+ppty+'...')
        self._checks_result = {psn: False for psn in psnames}
        msg = 'Successfully checked '+ppty+' preparation for {}/' + \
            str(len(psnames))
        checked = 0
        time = _time.time()
        while _time.time() - time < timeout:
            for psname in psnames:
                if self._checks_result[psname]:
                    continue
                cycler = self._get_cycler(psname)
                if ppty == 'parameters':
                    ret = cycler.is_prepared(self.mode, 0.05)
                elif ppty == 'opmode':
                    ret = cycler.check_opmode_cycle(self.mode, 0.05)
                if ret:
                    self._checks_result[psname] = True
                    checked = sum(self._checks_result.values())
                    if not checked % 5:
                        self._update_log(msg.format(str(checked)))
                if _time.time() - time > timeout:
                    break
            if all(self._checks_result.values()):
                break
            _time.sleep(TIMEOUT_SLEEP)
        self._update_log(msg.format(str(checked)))

        status = True
        for psname, sts in self._checks_result.items():
            if sts:
                continue
            self._update_log(psname+' is not ready.', error=True)
            status &= False
        return status

    def check_timing(self):
        """Check timing preparation."""
        if self._not_ctrl_ti:
            return True

        self._update_log('Checking Timing...')
        time0 = _time.time()
        while _time.time() - time0 < 5:
            status = self._timing.check(self.mode, self._triggers)
            if status:
                self._update_log(done=True)
                return True
            _time.sleep(TIMEOUT_SLEEP)
        self._update_log(
            'Timing is not configured or has problems. Please verify.',
            error=True)
        return False

    def check_egun_off(self):
        """Check egun off."""
        if 'LI-01:PS-Spect' in self.psnames:
            status = (self._pv_egun.value == 0)
            if not status:
                self._update_log(
                    'Linac EGun pulse is enabled! '
                    'Please disable it.', error=True)
            return status
        return True

    def set_triggers_state(self, triggers, state):
        """Set triggers state."""
        if self._not_ctrl_ti:
            return True

        label = 'enabl' if state == 'enbl' else 'disabl'
        self._update_log(label.capitalize()+'ing triggers...')
        if self._timing.set_triggers_state(state, triggers):
            self._update_log(done=True)
            return True
        self._update_log('Could not '+label+'e triggers!', error=True)
        return False

    def trigger_timing(self):
        """Trigger timing according to mode."""
        if self._not_ctrl_ti:
            return
        self._update_log('Triggering timing...')
        self._timing.trigger(self.mode)
        self._update_log(done=True)

    def init_trims(self, trims):
        """Initialize trims cycling process."""
        # initialize dict to check which trim is cycling
        self._is_cycling_dict = {ps: True for ps in trims}
        # trigger
        self.trigger_timing()

    def wait_trims(self):
        """Wait Trims to cycle."""
        self._update_log('Waiting for trims to cycle...')
        time0 = _time.time()
        keep_waiting = True
        while keep_waiting:
            # update remaining time
            _time.sleep(1)
            time = round(self._cycle_trims_duration - (_time.time()-time0))
            self._update_log('Remaining time: {}s...'.format(time))

            # verify if trims started to cycle
            if 14 < _time.time() - time0 < 15:
                for psname in self._is_cycling_dict:
                    cycler = self._get_cycler(psname)
                    if not cycler.get_cycle_enable():
                        self._update_log(
                            psname + ' is not cycling!', warning=True)
                        self._is_cycling_dict[psname] = False
                if sum(self._is_cycling_dict.values()) == 0:
                    self._update_log(
                        'All trims failed. Stopping.', error=True)
                    return False

            # update keep_waiting
            keep_waiting = _time.time() - time0 < self._cycle_trims_duration

        self._update_log(done=True)
        return True

    def cycle_trims_subset(self, trims, timeout=TIMEOUT_CHECK):
        """Cycle trims."""
        if not self.check_pwrsupplies('parameters', trims, timeout):
            return False

        self.config_pwrsupplies('opmode', trims)
        if not self.check_pwrsupplies('opmode', trims, timeout):
            return False

        triggers = _get_trigger_by_psname(trims)
        if not self.set_triggers_state(triggers, 'enbl'):
            return False
        self.init_trims(trims)
        if not self.wait_trims():
            self.set_pwrsupplies_slowref(trims)
            return False
        if not self.set_triggers_state(triggers, 'dsbl'):
            self.set_pwrsupplies_slowref(trims)
            return False

        if not self.check_pwrsupplies_finalsts(trims):
            self.set_pwrsupplies_slowref(trims)
            return False

        self.set_pwrsupplies_slowref(trims)
        if not self.check_pwrsupplies_slowref(trims, timeout):
            return False
        return True

    def init(self):
        """Initialize cycling process."""
        # initialize dict to check which ps is cycling
        self._is_cycling_dict = {ps: True for ps in self.psnames}

        self._aux_threads = list()
        psn2thr = [
            p for p in self.psnames if _PVName(p).sec == 'LI' or
            _PVName(p).dev in ['FCH', 'FCV']]
        for psname in psn2thr:
            cycler = self._get_cycler(psname)
            thread = _thread.Thread(target=cycler.cycle, daemon=True)
            self._aux_threads.append(thread)
            thread.start()

        # trigger
        self.trigger_timing()

    def wait(self):
        """Wait/Sleep while cycling according to mode."""
        self._update_log('Waiting for cycling...')
        time0 = _time.time()
        keep_waiting = True
        while keep_waiting:
            # update remaining time
            _time.sleep(1)
            if self.mode == 'Cycle':
                time = round(self._cycle_duration - (_time.time()-time0))
            else:
                time = round(
                    self._cycle_duration -
                    self._timing.get_cycle_count() *
                    DEFAULT_RAMP_DURATION/1000000)
            self._update_log('Remaining time: {}s...'.format(time))

            # verify if power supplies started to cycle
            if (self.mode == 'Cycle') and (5 < _time.time() - time0 < 6):
                for psname in self.psnames:
                    if _PVName(psname).sec == 'LI':
                        continue
                    if _PVName(psname).dev in ['FCH', 'FCV']:
                        continue
                    cycler = self._get_cycler(psname)
                    if not cycler.get_cycle_enable():
                        self._update_log(
                            psname + ' is not cycling!', warning=True)
                        self._is_cycling_dict[psname] = False
                if sum(self._is_cycling_dict.values()) == 0:
                    self._update_log(
                        'All power supplies failed. Stopping.', error=True)
                    return False

            # update keep_waiting
            if self.mode == 'Cycle':
                keep_waiting = _time.time() - time0 < self._cycle_duration
            else:
                keep_waiting = not self._timing.check_ramp_end()

        self._update_log(done=True)
        return True

    def check_pwrsupplies_finalsts(self, psnames):
        """Check all power supplies final state according to mode."""
        self._update_log('Checking power supplies final state...')
        self._checks_final_result = dict()
        sucess_cnt = 0
        msg = 'Successfully checked final state for {}/'+str(len(psnames))
        for psname in psnames:
            cycler = self._get_cycler(psname)
            status = cycler.check_final_state(self.mode)
            if status == _Const.CycleEndStatus.Ok:
                sucess_cnt += 1
                if sucess_cnt % 5 == 0:
                    self._update_log(msg.format(str(sucess_cnt)))
            self._checks_final_result[psname] = status
        self._update_log(msg.format(str(sucess_cnt)))

        all_ok = True
        for psname, has_prob in self._checks_final_result.items():
            if has_prob == _Const.CycleEndStatus.LackTriggers:
                self._update_log(
                    'Verify the number of pulses '+psname+' received!',
                    error=True)
                all_ok = False
            elif has_prob == _Const.CycleEndStatus.NotFinished \
                    and self._is_cycling_dict[psname]:
                self._update_log(
                    psname+' is finishing cycling...', warning=True)
            elif has_prob == _Const.CycleEndStatus.Interlock:
                self._update_log(
                    psname+' has interlock problems.', error=True)
                all_ok = False
        return all_ok

    def set_pwrsupplies_sofbmode(self, psnames):
        """Set power supplies SOFBMode."""
        psnames = {
            p for p in psnames if _PSSearch.conv_psname_2_psmodel(p) == 'FBP'}
        if not psnames:
            return

        self._update_log('Turning off power supplies SOFBMode...')
        for idx, psname in enumerate(psnames):
            cycler = self._get_cycler(psname)
            cycler.set_sofbmode('off')
            if idx % 5 == 4 or idx == len(psnames)-1:
                self._update_log(
                    'Sent SOFBMode preparation to {0}/{1}'.format(
                        str(idx+1), str(len(psnames))))

    def check_pwrsupplies_sofbmode(self, psnames, timeout=TIMEOUT_CHECK):
        """Check power supplies SOFBMode."""
        psnames = {
            p for p in psnames if _PSSearch.conv_psname_2_psmodel(p) == 'FBP'}
        if not psnames:
            return True

        self._update_log('Checking power supplies SOFBMode...')
        self._checks_result = {psn: False for psn in psnames}
        msg = 'Successfully checked SOFBMode preparation for {}/' + \
            str(len(psnames))
        checked = 0
        time = _time.time()
        while _time.time() - time < timeout:
            for psname in psnames:
                if self._checks_result[psname]:
                    continue
                cycler = self._get_cycler(psname)
                if cycler.check_sofbmode('off', 0.05):
                    self._checks_result[psname] = True
                    checked = sum(self._checks_result.values())
                    if not checked % 5:
                        self._update_log(msg.format(str(checked)))
                if _time.time() - time > timeout:
                    break
            if all(self._checks_result.values()):
                break
            _time.sleep(TIMEOUT_SLEEP)
        self._update_log(msg.format(str(checked)))

        status = True
        for psname, sts in self._checks_result.items():
            if sts:
                continue
            self._update_log(psname+' is in SOFBMode.', error=True)
            status &= False
        return status

    def set_pwrsupplies_slowref(self, psnames):
        """Set power supplies OpMode to SlowRef."""
        psnames = {p for p in psnames if 'LI' not in p}
        if not psnames:
            return

        self._update_log('Setting power supplies to SlowRef...')
        with ThreadPoolExecutor(max_workers=100) as executor:
            for idx, psname in enumerate(psnames):
                cycler = self._get_cycler(psname)
                target = cycler.set_opmode_slowref
                executor.submit(target)
                if idx % 5 == 4 or idx == len(psnames)-1:
                    self._update_log(
                        'Sent opmode preparation to {0}/{1}'.format(
                            str(idx+1), str(len(psnames))))

    def check_pwrsupplies_slowref(self, psnames, timeout=TIMEOUT_CHECK):
        """Check power supplies OpMode."""
        psnames = {p for p in psnames if 'LI' not in p}
        if not psnames:
            return True

        self._update_log('Checking power supplies opmode...')
        self._checks_result = {psn: False for psn in psnames}
        msg = 'Successfully checked opmode preparation for {}/' + \
            str(len(psnames))
        checked = 0
        time = _time.time()
        while _time.time() - time < timeout:
            for psname in psnames:
                if self._checks_result[psname]:
                    continue
                cycler = self._get_cycler(psname)
                if cycler.check_opmode_slowref(0.05):
                    self._checks_result[psname] = True
                    checked = sum(self._checks_result.values())
                    if not checked % 5:
                        self._update_log(msg.format(str(checked)))
                if _time.time() - time > timeout:
                    break
            if all(self._checks_result.values()):
                break
            _time.sleep(TIMEOUT_SLEEP)
        self._update_log(msg.format(str(checked)))

        status = True
        for psname, sts in self._checks_result.items():
            if sts:
                continue
            opmdes = 'closed_loop_manual' if 'FC' in psname else 'SlowRef'
            self._update_log(psname+' is not in '+opmdes+'.', error=True)
            status &= False
        return status

    def set_pwrsupplies_current_zero(self, psnames):
        """Set power supplies current to zero."""
        self._update_log('Setting power supplies current to zero...')
        for idx, psname in enumerate(psnames):
            cycler = self._get_cycler(psname)
            cycler.set_current_zero()
            if idx % 5 == 4 or idx == len(psnames)-1:
                self._update_log(
                    'Sent current preparation to {0}/{1}'.format(
                        str(idx+1), str(len(psnames))))

    def check_pwrsupplies_current_zero(self, psnames, timeout=TIMEOUT_CHECK):
        """Check power supplies current."""
        self._update_log('Checking power supplies current...')
        self._checks_result = {psn: False for psn in psnames}
        msg = 'Successfully checked current preparation for {}/' + \
            str(len(psnames))
        checked = 0
        time = _time.time()
        while _time.time() - time < timeout:
            for psname in psnames:
                if self._checks_result[psname]:
                    continue
                cycler = self._get_cycler(psname)
                if cycler.check_current_zero(0.05):
                    self._checks_result[psname] = True
                    checked = sum(self._checks_result.values())
                    if not checked % 5:
                        self._update_log(msg.format(str(checked)))
                if _time.time() - time > timeout:
                    break
            if all(self._checks_result.values()):
                break
            _time.sleep(TIMEOUT_SLEEP)
        self._update_log(msg.format(str(checked)))

        status = True
        for psname, sts in self._checks_result.items():
            if sts:
                continue
            self._update_log(psname+' current is not zero.', error=True)
            status &= False
        return status

    # --- main commands ---

    def save_timing_initial_state(self):
        """Save timing initial state."""
        self._update_log('Saving Timing initial state...')
        self._timing.save_initial_state()
        self._update_log('...finished!')

    def prepare_timing(self):
        """Prepare Timing."""
        self.config_timing()
        if not self.check_timing():
            return
        self._update_log('Timing preparation finished!')

    def prepare_pwrsupplies_sofbmode(self):
        """Prepare SOFBMode."""
        psnames = self.psnames
        timeout = TIMEOUT_CHECK
        if self._include_sitrims:
            self.create_trims_cyclers()
            psnames.extend(self.trimnames)
            timeout += TIMEOUT_CHECK

        self.set_pwrsupplies_sofbmode(psnames)
        if not self.check_pwrsupplies_sofbmode(psnames, timeout):
            self._update_log(
                'There are power supplies in SOFBMode.', error=True)
            return
        self._update_log('Power supplies SOFBMode preparation finished!')

    def prepare_pwrsupplies_opmode_slowref(self):
        """Prepare OpMode to slowref."""
        psnames = self.psnames
        timeout = TIMEOUT_CHECK
        if self._include_sitrims:
            self.create_trims_cyclers()
            psnames.extend(self.trimnames)
            timeout += 3*TIMEOUT_CHECK

        self.set_pwrsupplies_slowref(psnames)
        if not self.check_pwrsupplies_slowref(psnames, timeout):
            self._update_log(
                'There are power supplies not in '
                'the correct OpMode.', error=True)
            return
        self._update_log('Power supplies OpMode preparation finished!')

    def prepare_pwrsupplies_current_zero(self):
        """Prepare current to cycle."""
        psnames = self.psnames
        timeout = TIMEOUT_CHECK
        if self._include_sitrims:
            self.create_trims_cyclers()
            psnames.extend(self.trimnames)
            timeout += 3*TIMEOUT_CHECK

        self.set_pwrsupplies_current_zero(psnames)
        if not self.check_pwrsupplies_current_zero(psnames, timeout):
            self._update_log(
                'There are power supplies with current not zero.', error=True)
            return
        self._update_log('Power supplies current preparation finished!')

    def prepare_pwrsupplies_parameters(self):
        """Prepare parameters to cycle."""
        psnames = self.psnames
        timeout = TIMEOUT_CHECK
        if self._include_sitrims:
            self.create_trims_cyclers()
            psnames.extend(self.trimnames)
            timeout += 3*TIMEOUT_CHECK

        self.config_pwrsupplies('parameters', psnames)
        if not self.check_pwrsupplies('parameters', psnames, timeout):
            self._update_log(
                'There are power supplies not configured to cycle.',
                error=True)
            return
        self._update_log('Power supplies parameters preparation finished!')

    def prepare_pwrsupplies_opmode_cycle(self):
        """Prepare OpMode to cycle."""
        psnames = self.psnames
        self.config_pwrsupplies('opmode', psnames)
        if not self.check_pwrsupplies('opmode', psnames):
            self._update_log(
                'There are power supplies with wrong opmode.', error=True)
            return
        self._update_log('Power supplies OpMode preparation finished!')

    def cycle_trims(self):
        """Cycle all trims."""
        if not self._include_sitrims:
            return

        if not self.check_timing():
            return

        self.create_trims_cyclers()

        self._update_log('Preparing to cycle CHs, QSs and QTrims...')
        trims = _PSSearch.get_psnames({
            'sec': 'SI', 'sub': '[0-2][0-9](M|C).*', 'dis': 'PS',
            'dev': '(CH|QS|QD.*|QF.*|Q[1-4])'})
        if not self.cycle_trims_subset(trims, timeout=50):
            self._update_log(
                'There was problems in trims cycling. Stoping.', error=True)
            return

        self._update_log('Preparing to cycle CVs...')
        trims = _PSSearch.get_psnames({
            'sec': 'SI', 'sub': '[0-2][0-9](M|C).*', 'dis': 'PS',
            'dev': 'CV'})
        if not self.cycle_trims_subset(trims, timeout=50):
            self._update_log(
                'There was problems in trims cycling. Stoping.', error=True)
            return

        self._update_log('Trims cycle finished!')

    def cycle(self):
        """Cycle."""
        # check
        if not self.check_egun_off():
            return
        if not self.check_timing():
            return

        if not self.check_pwrsupplies('parameters', self.psnames):
            self._update_log(
                'There are power supplies not configured to cycle. Stopping.',
                error=True)
            return
        if not self.check_pwrsupplies('opmode', self.psnames):
            self._update_log(
                'There are power supplies with wrong opmode. Stopping.',
                error=True)
            return

        triggers = self._triggers.copy()
        if self._include_sitrims:
            triggers.difference_update(self._si_aux_triggers)
        if not self.set_triggers_state(triggers, 'enbl'):
            return
        self.init()
        if not self.wait():
            return

        self.check_pwrsupplies_finalsts(self.psnames)
        self.set_pwrsupplies_slowref(self.psnames)
        if not self.check_pwrsupplies_slowref(self.psnames):
            return False

        # Indicate cycle end
        self._update_log('Cycle finished!')

    def restore_timing_initial_state(self):
        """Restore timing initial state."""
        self._update_log('Restoring Timing initial state...')
        self._timing.restore_initial_state()
        self._update_log('...finished!')

    # --- private methods ---

    def _filter_psnames(self, psnames2filt, filt):
        if psnames2filt:
            return _Filter.process_filters(psnames2filt, filters=filt)
        return _PSSearch.get_psnames(filt)

    def _get_cycler(self, psname):
        if psname in self._cyclers:
            return self._cyclers[psname]
        if psname in self._aux_cyclers:
            return self._aux_cyclers[psname]
        raise ValueError('There is no cycler defined to '+psname+'!')

    def _update_log(self, message='', done=False, warning=False, error=False):
        self._logger_message = message
        if self._logger:
            self._logger.update(message, done, warning, error)
        else:
            if done and not message:
                message = 'Done.'
            _log.info(message)



#!/usr/bin/env python-sirius

"""Module to test CurrInfo Lifetime Soft IOC main module."""

import unittest
from unittest import mock
import siriuspy.util as util
from siriuspy.currinfo import SILifetimeApp


PUB_INTERFACE = (
    'pvs_database',
    'init_database',
    'process',
    'read',
    'write',
)


class TestASAPCurrInfoLifetimeMain(unittest.TestCase):
    """Test AS-AP-CurrInfo Lifetime Soft IOC."""

    def setUp(self):
        pv_patcher = mock.patch(
            "siriuspy.currinfo.lifetime.main._PV", autospec=True)
        self.addCleanup(pv_patcher.stop)
        self.mock_pv = pv_patcher.start()
        self.mock_pv.return_value.connected = False
        self.app = SILifetimeApp()

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            SILifetimeApp, PUB_INTERFACE, print_flag=True)
        self.assertTrue(valid)

    def test_write_MinIntvlBtwSpl(self):
        """Test write MinIntvlBtwSmpl-SP."""
        self.app.write('MinIntvlBtwSpl-SP', -1)
        self.assertEqual(self.app._min_intvl_btw_spl, 0)

        self.app.write('MinIntvlBtwSpl-SP', 0)
        self.assertEqual(self.app._min_intvl_btw_spl, 0)

        self.app.write('MinIntvlBtwSpl-SP', 1)
        self.assertEqual(self.app._min_intvl_btw_spl, 1)

    def test_write_SplIntvl(self):
        """Test write MaxSplIntvl-SP."""
        self.app.write('MaxSplIntvl-SP', 100)
        self.assertEqual(self.app._sampling_interval, 100)

    def test_write_LtFitMode(self):
        """Test write LtFitMode-Sel."""
        self.app.write('LtFitMode-Sel', 1)
        self.assertEqual(self.app._mode, 1)

    def test_write_CurrOffset(self):
        """Test write CurrOffset-SP."""
        self.app.write('CurrOffset-SP', 1)
        self.assertEqual(self.app._current_offset, 1)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python-sirius

"""Module to test AS-AP-TuneCorr Soft IOC main module."""

import unittest
from unittest import mock
import siriuspy.util as util
from siriuspy.opticscorr.csdev import Const
from siriuspy.opticscorr.tune import TuneCorrApp


PUB_INTERFACE = (
    'update_corrparams_pvs',
    'set_dtune_x',
    'set_dtune_y',
    'cmd_set_newref',
    'set_meas_config_dkl_qf',
    'set_meas_config_dkl_qd',
)


class TestASAPTuneCorrMain(unittest.TestCase):
    """Test AS-AP-TuneCorr Soft IOC."""

    def setUp(self):
        """Initialize Soft IOC."""
        self.q_ok = {
            'matrix': [
                [2.7280, 8.5894, 4.2995, 0.5377,
                 1.0906, 2.0004, 0.5460, 1.0012],
                [-1.3651, -3.5532, -1.7657, -2.3652,
                 -4.7518, -1.9781, -2.3601, -0.9839]],
            'nominal KLs': [0.7146, 1.2344, 1.2344, -0.2270,
                            -0.2809, -0.4783, -0.2809, -0.4783]}
        self.qfams = ['QFA', 'QFB', 'QFP',
                      'QDA', 'QDB1', 'QDB2', 'QDP1', 'QDP2']
        cs_patcher = mock.patch(
            "siriuspy.opticscorr.base._ConfigDBClient", autospec=True)
        self.addCleanup(cs_patcher.stop)
        self.mock_cs = cs_patcher.start()
        self.mock_cs.return_value.get_config_value.return_value = self.q_ok
        pv_patcher = mock.patch(
            "siriuspy.opticscorr.base._PV", autospec=True)
        self.addCleanup(pv_patcher.stop)
        self.mock_pv = pv_patcher.start()
        cnh_patcher = mock.patch(
            "siriuspy.opticscorr.base._HandleConfigNameFile",
            autospec=True)
        self.addCleanup(cnh_patcher.stop)
        self.mock_cnh = cnh_patcher.start()
        self.mock_cnh.return_value.get_config_name.return_value = \
            'SI.V24.04_S05.01'
        self.app = TuneCorrApp('SI')

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(
            TuneCorrApp, PUB_INTERFACE, print_flag=True)
        self.assertTrue(valid)

    def test_write_ok_syncoff_Apply(self):
        """Test write on ApplyDelta-Cmd in normal operation, sync off."""
        self.app._sync_corr = Const.SyncCorr.Off

        self.app._status = 0b00000
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, len(self.qfams))

        self.app._status = 0b10000
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 2*len(self.qfams))

    def test_write_ok_syncon_Apply(self):
        """Test write on ApplyDelta-Cmd in normal operation, sync on."""
        self.app._sync_corr = Const.SyncCorr.On
        self.app._status = 0b00000
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 1+len(self.qfams))

    def test_write_err_syncon_Apply(self):
        """Test write on ApplyDelta-Cmd on status error."""
        self.app._sync_corr = Const.SyncCorr.On
        self.app._status = 0b10000
        self.assertFalse(self.app.write('ApplyDelta-Cmd', 0))
        self.mock_pv.return_value.put.assert_not_called()

    def test_write_CorrMeth(self):
        """Test write on CorrMeth-Sel."""
        self.assertTrue(self.app.write('CorrMeth-Sel', 1))
        self.assertEqual(self.app._corr_method, 1)

    def test_write_SyncCorr(self):
        """Test write on SyncCorr-Sel."""
        self.mock_pv.return_value.connected = True
        self.assertTrue(self.app.write('SyncCorr-Sel', 1))
        self.assertEqual(self.app._sync_corr, 1)

    def test_write_ok_ConfigPS(self):
        """Test write on ConfigPS-Cmd in normal operation."""
        self.mock_pv.return_value.connected = True
        self.assertFalse(self.app.write('ConfigPS-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 2*len(self.qfams))

    def test_write_err_ConfigPS(self):
        """Test write on ConfigPS-Cmd on connection error."""
        self.mock_pv.return_value.connected = False
        self.assertFalse(self.app.write('ConfigPS-Cmd', 0))
        self.mock_pv.return_value.put.assert_not_called()

    def test_write_ok_ConfigTI(self):
        """Test write on ConfigTiming-Cmd in normal operation."""
        self.mock_pv.return_value.connected = True
        self.assertFalse(self.app.write('ConfigTiming-Cmd', 0))
        count = self.mock_pv.return_value.put.call_count
        self.assertEqual(count, 9)

    def test_write_err_ConfigTI(self):
        """Test write on ConfigTiming-Cmd in connection error."""
        self.mock_pv.return_value.connected = False
        self.assertFalse(self.app.write('ConfigTiming-Cmd', 0))
        self.mock_pv.return_value.put.assert_not_called()

    def test_write_ok_SetNewRefKL(self):
        """Test write on SetNewRefKL-Cmd in normal operation."""
        self.mock_pv.return_value.get.return_value = 0.0
        self.app._status = 0
        self.assertFalse(self.app.write('SetNewRefKL-Cmd', 0))
        self.assertEqual(self.app._delta_tunex, 0)
        self.assertEqual(self.app._delta_tuney, 0)
        for fam in self.qfams:
            self.assertEqual(self.app._lastcalc_deltakl[fam], 0)

    def test_write_err_SetNewRefKL(self):
        """Test write on SetNewRefKL-Cmd on connection error."""
        self.app._status = 0b00001
        self.assertFalse(self.app.write('SetNewRefKL-Cmd', 0))
        self.mock_pv.return_value.get.assert_not_called()


if __name__ == "__main__":
    unittest.main()



#!/usr/local/bin/python-sirius
"""Driver module."""

from pcaspy import Alarm as _Alarm, Severity as _Severity

from ...namesys import SiriusPVName
from ..app import App as _App
from ..pvs import ComputedPV as _ComputedPV
from .pvs import PUStatusPV as _PUStatusPV, PUDiffPV as _PUDiffPV


class PUDiagApp(_App):
    """Main application responsible for updating DB."""

    def _create_computed_pvs(self, punames):
        self._punames = punames
        for puname in self._punames:
            devname = SiriusPVName(puname).substitute(prefix=self._prefix)

            # DiagVoltageDiff-Mon
            pvs = [None, None]
            pvs[_PUDiffPV.VOLTAGE_SP] = devname + ':Voltage-SP'
            pvs[_PUDiffPV.VOLTAGE_MON] = devname + ':Voltage-Mon'
            pv = _ComputedPV(
                puname + ':DiagVoltageDiff-Mon', _PUDiffPV(), self._queue,
                pvs, monitor=False)
            self.pvs.append(pv)

            # DiagStatus-Mon
            pvs = [None]*11 if 'Sept' not in puname else [None]*10
            pvs[_PUStatusPV.PWRST_STS] = devname + ':PwrState-Sts'
            pvs[_PUStatusPV.PULSE_STS] = devname + ':Pulse-Sts'
            pvs[_PUStatusPV.DIFFVOLTG] = devname + ':DiagVoltageDiff-Mon'
            pvs[_PUStatusPV.INTRLCK_1] = devname + ':Intlk1-Mon'
            pvs[_PUStatusPV.INTRLCK_2] = devname + ':Intlk2-Mon'
            pvs[_PUStatusPV.INTRLCK_3] = devname + ':Intlk3-Mon'
            pvs[_PUStatusPV.INTRLCK_4] = devname + ':Intlk4-Mon'
            pvs[_PUStatusPV.INTRLCK_5] = devname + ':Intlk5-Mon'
            pvs[_PUStatusPV.INTRLCK_6] = devname + ':Intlk6-Mon'
            pvs[_PUStatusPV.INTRLCK_7] = devname + ':Intlk7-Mon'
            if 'Sept' not in puname:
                pvs[_PUStatusPV.INTRLCK_8] = devname + ':Intlk8-Mon'
            pv = _ComputedPV(
                puname + ':DiagStatus-Mon', _PUStatusPV(), self._queue,
                pvs, monitor=False)
            self.pvs.append(pv)
        self._pvs_connected = {pv: False for pv in self.pvs}

    def _update_pvs(self):
        for pvo in self.pvs:
            if not pvo.connected:
                if self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.TIMEOUT_ALARM,
                        severity=_Severity.INVALID_ALARM,
                        field='status')
                self._pvs_connected[pvo] = False
                if 'DiagStatus' in pvo.pvname:
                    self.run_callbacks(pvo.pvname, value=pvo.value)
            else:
                if not self._pvs_connected[pvo]:
                    self.run_callbacks(
                        pvo.pvname, alarm=_Alarm.NO_ALARM,
                        severity=_Severity.NO_ALARM, field='status')
                self._pvs_connected[pvo] = True
                self.run_callbacks(pvo.pvname, value=pvo.value)



"""Control system Device Util Module."""

import copy as _copy

from mathphys.functions import get_namedtuple as _get_namedtuple

from . import util as _util
from .clientweb import beaglebone_ip_list as _bbb_ip_list
from .search import PSSearch as _PSSearch


_DEV_2_IOC_IP_DICT = dict()


class ETypes:
    """Enumerate types."""

    DSBL_ENBL = ('Dsbl', 'Enbl')
    DSBLD_ENBLD = ('Dsbld', 'Enbld')
    OFF_ON = ('Off', 'On')
    CLOSE_OPEN = ('Closed', 'Open')
    DISCONN_CONN = ('Disconnected', 'Connected')
    FIXED_INCR = ('Incr', 'Fixed')
    NORM_INV = ('Normal', 'Inverse')
    UNLINK_LINK = ('Unlink', 'Link')


_et = ETypes  # syntactic sugar


class Const:
    """Const class defining power supply constants."""

    DsblEnbl = _get_namedtuple('DsblEnbl', _et.DSBL_ENBL)
    OffOn = _get_namedtuple('OffOn', _et.OFF_ON)
    CloseOpen = _get_namedtuple('CloseOpen', _et.CLOSE_OPEN)
    DisconnConn = _get_namedtuple('DisconnConn', _et.DISCONN_CONN)

    @staticmethod
    def register(name, field_names, values=None):
        """Register namedtuple."""
        return _get_namedtuple(name, field_names, values)


def add_pvslist_cte(database, prefix=''):
    """Add Properties-Cte."""
    pvslist_cte_name = prefix + 'Properties-Cte'
    keys = list(database.keys())
    keys.append(pvslist_cte_name)
    val = ' '.join(sorted(keys))
    database[pvslist_cte_name] = {
        'type': 'char',
        'count': len(val),
        'value': val,
    }
    return database


def get_device_2_ioc_ip(reload=False):
    """Return a dict of ioc IP numbers for csdevices."""
    if not _DEV_2_IOC_IP_DICT or reload:
        _reload_device_2_ioc_ip()
    return _copy.deepcopy(_DEV_2_IOC_IP_DICT)


def _reload_device_2_ioc_ip():

    # beaglebone IPs
    text, _ = _util.read_text_data(_bbb_ip_list())
    for item in text:
        if len(item) == 2:
            bbbname, ipm = item
            _DEV_2_IOC_IP_DICT[bbbname] = ipm

    # power supplies
    dic = dict()
    for bbbname, ipm in _DEV_2_IOC_IP_DICT.items():
        try:
            bsmps = _PSSearch.conv_bbbname_2_bsmps(bbbname)
            for bsmp in bsmps:
                psname, _ = bsmp
                dic[psname] = ipm
        except KeyError:
            pass
    _DEV_2_IOC_IP_DICT.update(dic)



"""Definition of factories.

NormalizerFactory
    used to create magnet normalizers
"""

from ..search import PSSearch as _PSSearch
from ..search import MASearch as _MASearch

from . import util as _mutil
from . import normalizer as _norm


class NormalizerFactory:
    """Factory class for normalizer objects."""

    @staticmethod
    def create(maname):
        """Return appropriate normalizer."""
        psname = _MASearch.conv_maname_2_psnames(maname)[0]
        magfunc = _PSSearch.conv_psname_2_magfunc(psname)
        ma_class = _mutil.magnet_class(maname)
        if ma_class == 'dipole':
            return _norm.DipoleNormalizer(maname, magnet_conv_sign=-1.0)
        if ma_class == 'trim':
            return _norm.TrimNormalizer(maname, magnet_conv_sign=-1.0)
        if ma_class == 'id-apu':
            return _norm.APUNormalizer(maname, magnet_conv_sign=+1.0)
        if magfunc == 'corrector-horizontal':
            return _norm.MagnetNormalizer(maname, magnet_conv_sign=+1.0)
        if 'TB' in maname and 'QD' in maname:
            return _norm.MagnetNormalizer(maname, magnet_conv_sign=+1.0,)
        return _norm.MagnetNormalizer(maname, magnet_conv_sign=-1.0)



#!/usr/local/bin/python-sirius
"""Computed PV."""

import numpy as _np

from ..epics import PV as _PV, CONNECTION_TIMEOUT as _CONN_TIMEOUT


class ComputedPV:
    """Computed PV class.

    Objects of this class are used for properties or process variables that are
    computed from other primary process variables. magnet strengths which
    are derived from power supply currents are typical examples of such
    computed process variables.
    """

    def __init__(self, pvname, computer, queue, pvs, monitor=True):
        """Initialize PVs."""
        # print('compute_pv: ', pvname, pvs)

        # starts computer_pvs queue, if not started yet
        self._queue = queue
        if not self._queue.running:
            self._queue.start()

        # --- properties ---

        self.pvname = pvname
        self._value = None
        self._set_limits((None,)*6)
        self.computer = computer
        self.pvs = self._create_primary_pvs_list(pvs)

        # flags if computed_pv is of the 'monitor' type.
        self._monitor_pv = \
            '-Mon' in self.pvname and 'Ref-Mon' not in self.pvname

        # add callback
        self._callbacks = {}
        self._monitor = monitor
        if self._monitor:
            if self._monitor_pv:
                # in order to optimize efficiency, if computed pv is of the
                # monitor type add callback only to the first primary pv, the
                # one corresponding to the main current.
                self.pvs[0].add_callback(self._value_update_callback)
            else:
                for pvobj in self.pvs:
                    pvobj.add_callback(self._value_update_callback)

        # init limits
        if self.connected:
            lims = self.computer.compute_limits(self)
            self._set_limits(lims)

        for pvobj in self.pvs:
            pvobj.run_callbacks()

    # --- public methods ---

    @property
    def connected(self):
        """Return wether all pvs are connected."""
        for pvobj in self.pvs:
            if not pvobj.connected:
                # print(pvobj.pvname)
                return False
        return True

    @property
    def value(self):
        """Return computed PV value."""
        return self.get()
        # return self._value

    def get(self):
        """Return current value of computed PV."""
        if self._monitor:
            pass
        else:
            self._update_value()
        return self._value

    def put(self, value):
        """Put `value` to the first pv of the pv list."""
        self._value = value
        self.computer.compute_put(self, value)

    def add_callback(self, func, index=None):
        """Add callback to be issued when a PV is updated."""
        if not callable(func):
            raise ValueError("Tried to set non callable as a callback")
        if index is None:
            index = 0 if len(self._callbacks) == 0 \
                else max(self._callbacks.keys()) + 1
        self._callbacks[index] = func
        return len(self._callbacks) - 1

    def run_callbacks(self):
        """Run all callbacks."""
        self._issue_callback(**{
            'pvname': self.pvname,
            'value': self._value,
            'hihi': self.upper_alarm_limit,
            'high': self.upper_warning_limit,
            'hilim': self.upper_disp_limit,
            'lolim': self.lower_disp_limit,
            'low': self.lower_warning_limit,
            'lolo': self.lower_alarm_limit,
        })

    def wait_for_connection(self, timeout):
        """Wait util computed PV is connected or until timeout."""
        pass

    # --- private methods ---

    def _set_limits(self, lims):
        self.upper_alarm_limit = lims[0]
        self.upper_warning_limit = lims[1]
        self.upper_disp_limit = lims[2]
        self.lower_disp_limit = lims[3]
        self.lower_warning_limit = lims[4]
        self.lower_alarm_limit = lims[5]

    def _create_primary_pvs_list(self, pvs):
        # get list of primary pvs
        ppvs = list()  # List with PVs used by the computed PV
        for pvobj in pvs:
            if isinstance(pvobj, str):  # give up string option.
                tpv = _PV(pvobj, connection_timeout=_CONN_TIMEOUT)
                ppvs.append(tpv)
            else:
                ppvs.append(pvobj)
        return ppvs

    def _is_same(self, value):
        """."""
        # make sure vars are converted to numpy array, if it is the case
        # TODO: maybe this checking can be simplified. the class seems to be
        # converting values unnecessarily.
        if isinstance(self._value, _np.ndarray) or isinstance(value, _np.ndarray):
            if not isinstance(self._value, _np.ndarray):
                self._value = _np.array(self._value)
            if not isinstance(value, _np.ndarray):
                value = _np.array(value)

        if isinstance(value, _np.ndarray):
            return _np.all(self._value == value)
        else:
            return self._value == value

    def _process_new_value(self, kwargs):
        self._value = kwargs["value"]
        # Check if limits are in the return dict and update them
        if "high" in kwargs:
            self.upper_alarm_limit = kwargs["hihi"]
            self.upper_warning_limit = kwargs["high"]
            self.upper_disp_limit = kwargs["hilim"]
            self.lower_disp_limit = kwargs["lolim"]
            self.lower_warning_limit = kwargs["low"]
            self.lower_alarm_limit = kwargs["lolo"]
        self._issue_callback(pvname=self.pvname, **kwargs)

    def _update_value(self, pvname=None, value=None):
        # Get dict with pv props that changed
        kwargs = self.computer.compute_update(self, pvname, value)

        if kwargs is None:
            return None

        if self._value is None:
            self._process_new_value(kwargs)
            return None

        if not self._is_same(kwargs['value']):
            self._process_new_value(kwargs)
            return None

        if 'high' not in kwargs:
            return None
        elif kwargs['hihi'] != self.upper_alarm_limit or \
                kwargs['high'] != self.upper_warning_limit or \
                kwargs['hilim'] != self.upper_disp_limit or \
                kwargs['hilim'] != self.lower_disp_limit or \
                kwargs['low'] != self.lower_warning_limit or \
                kwargs['lolo'] != self.lower_alarm_limit:
            self._process_new_value(kwargs)

    def _value_update_callback(self, pvname, value, **kwargs):
        # if 'Current-Mon' not in pvname:
        #     print(pvname, value)
        if self.connected:
            self._queue.add_callback(self._update_value, pvname, value)

    def _issue_callback(self, **kwargs):
        for callback in self._callbacks.values():
            callback(**kwargs)
